"""
 @file
 @brief This file loads the Initialize Effects / Pre-process effects dialog
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import time
import json
import functools
import webbrowser
import hashlib
import zipfile
from urllib.parse import urljoin

from qt_api import Qt, pyqtSignal, QCoreApplication, QTimer, QSize
from qt_api import QPainter
from qt_api import (
    QPushButton, QDialog, QLabel, QDoubleSpinBox, QSpinBox, QLineEdit,
    QCheckBox, QComboBox, QDialogButtonBox, QSizePolicy, QMessageBox,
    QFileDialog, QProgressDialog, QApplication, QWidget, QHBoxLayout,
)
import openshot  # Python module for libopenshot (required video editing module installed separately)

from classes import info
from classes import http_client
from classes import ui_util
from classes.app import get_app
from classes.logger import log
from classes.metrics import *

YOLO_MODELS_PATH = os.path.join(info.RESOURCES_PATH, "yolo-models.json")
EFFICIENT_SAM_MODELS_PATH = os.path.join(info.RESOURCES_PATH, "efficient-sam-models.json")
CUTIE_MODELS_PATH = os.path.join(info.RESOURCES_PATH, "cutie-models.json")
YOLO_MODEL_FILENAME = "model.onnx"
YOLO_CLASSES_FILENAME = "classes.names"
YOLO_INSTALL_METADATA = "install.json"
YOLO_FALLBACK_MODEL_NAME = "YOLO"
EFFICIENT_SAM_INSTALL_METADATA = "install-efficient-sam.json"
CUTIE_INSTALL_METADATA = "install-cutie.json"

EFFICIENT_SAM_MODEL_FILES = {
    "efficient-sam-tiny-1024": "image_segmentation_efficientsam_ti_2025april.onnx",
    "efficient-sam-small-static-1024": "image_segmentation_efficientsam_s_static_1024.onnx",
}

CUTIE_MODEL_FILES = {
    "cutie-low": {
        "encode-key": "cutie-encode-key-480x272.onnx",
        "encode-value": "cutie-encode-value-480x272.onnx",
        "memory-readout": "cutie-memory-readout-floatmask-valid-480x272-m6-topk30-opencv.onnx",
        "decode": "cutie-decode-480x272.onnx",
    },
    "cutie-medium": {
        "encode-key": "cutie-encode-key-640x368.onnx",
        "encode-value": "cutie-encode-value-640x368.onnx",
        "memory-readout": "cutie-memory-readout-floatmask-valid-640x368-m6-topk30-opencv.onnx",
        "decode": "cutie-decode-640x368.onnx",
    },
    "cutie-high": {
        "encode-key": "cutie-encode-key-960x544.onnx",
        "encode-value": "cutie-encode-value-960x544.onnx",
        "memory-readout": "cutie-memory-readout-floatmask-valid-960x544-m6-topk30-opencv.onnx",
        "decode": "cutie-decode-960x544.onnx",
    },
    "cutie-very-high": {
        "encode-key": "cutie-encode-key-1280x720.onnx",
        "encode-value": "cutie-encode-value-1280x720.onnx",
        "memory-readout": "cutie-memory-readout-floatmask-valid-1280x720-m6-topk30-opencv.onnx",
        "decode": "cutie-decode-1280x720.onnx",
    },
}


class DownloadCancelled(Exception):
    """Raised when a user cancels an in-progress download."""


def load_yolo_models_manifest():
    """Load the packaged allow-list of YOLO model downloads."""
    return load_model_manifest(YOLO_MODELS_PATH)


def load_model_manifest(path):
    """Load a packaged allow-list of model downloads."""
    with open(path, "r", encoding="utf-8") as manifest_file:
        manifest = json.load(manifest_file)
    manifest.setdefault("models", [])
    return manifest


def recommended_model(models):
    """Return the recommended model entry, or the first available entry."""
    for model in models:
        if model.get("recommended"):
            return model
    return models[0] if models else None


def yolo_model_dir(model):
    """Return the install directory for a packaged YOLO model entry."""
    model_id = model.get("id", "")
    if not model_id or os.path.basename(model_id) != model_id:
        raise ValueError("Invalid YOLO model id: %s" % model_id)
    return os.path.join(info.YOLO_PATH, model_id)


def model_install_dir(model):
    """Return the shared AI model install directory for a manifest entry."""
    return yolo_model_dir(model)


def yolo_model_path(model):
    """Return the installed ONNX path for a packaged YOLO model entry."""
    return os.path.join(yolo_model_dir(model), YOLO_MODEL_FILENAME)


def yolo_classes_path(model):
    """Return the installed class names path for a packaged YOLO model entry."""
    return os.path.join(yolo_model_dir(model), YOLO_CLASSES_FILENAME)


def yolo_model_label(model):
    """Return the dropdown label for a packaged YOLO model entry."""
    return model_label(model, YOLO_FALLBACK_MODEL_NAME)


def model_label(model, fallback_name="Model"):
    """Return the dropdown label for a packaged model entry."""
    label = model.get("name") or model.get("id") or fallback_name
    description = model.get("description")
    if description:
        label = "%s (%s)" % (label, description)
    return label


def translated_model_label(model, fallback_name="Model"):
    """Return a translated dropdown label for a packaged model entry."""
    _ = get_app()._tr
    label = _(model.get("name") or model.get("id") or fallback_name)
    description = model.get("description")
    if description:
        label = "%s (%s)" % (label, _(description))
    return label


def compact_model_label(model, fallback_name="Model"):
    """Return a compact dropdown label without descriptive prose."""
    return model.get("name") or model.get("id") or fallback_name


def object_mask_quality_label(model):
    """Return the Object Mask quality label for a Cutie manifest entry."""
    return model_label(model)


def yolo_download_button_label(model=None):
    """Return the download button label for a packaged YOLO model entry."""
    return model_download_button_label(model)


def model_download_button_label(model=None):
    """Return the download button label for a packaged model entry."""
    _ = get_app()._tr
    if not model or not model.get("bytes"):
        return _("Download")
    return _("Download (%s MB)") % max(1, int(round(model.get("bytes") / 1000000.0)))


class RegionButton(QPushButton):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.qimage = None

    def setImage(self, qimage):
        self.qimage = qimage
        self.update()  # Trigger a repaint

    def paintEvent(self, event):
        super().paintEvent(event)
        if self.qimage:
            painter = QPainter(self)
            resized_qimage = self.qimage.scaled(self.size(), Qt.IgnoreAspectRatio, Qt.SmoothTransformation)
            painter.drawImage(0, 0, resized_qimage)
        else:
            super().paintEvent(event)  # Draw the normal button


class ProcessEffect(QDialog):
    """ Choose Profile Dialog """
    progress = pyqtSignal(int)

    # Path to ui file
    ui_path = os.path.join(info.PATH, 'windows', 'ui', 'process-effect.ui')

    def __init__(self, clip_id, effect_class, effect_params):

        if not openshot.Clip().COMPILED_WITH_CV:
            raise ModuleNotFoundError("Openshot not compiled with OpenCV")

        # Create dialog class
        super().__init__()
        # Track effect details
        self.clip_id = clip_id
        self.effect_name = ""
        self.effect_class = effect_class
        self.context = {}
        self.file_fields = {}
        self.selection_fields = {}
        self.download_groups = []
        self.advanced_rows = []
        self.advanced_buttons = []
        self.advanced_visible = False
        self.onnx_validation_cache = {}
        self.yolo_models_manifest = None
        self.yolo_models = []
        self.model_manifests = {}
        self.processing_effect = False
        self.validation_wait_cursor = False
        self.file_validation_timer = QTimer(self)
        self.file_validation_timer.setInterval(300)
        self.file_validation_timer.setSingleShot(True)
        self.file_validation_timer.timeout.connect(self.update_file_validation)

        # Get all effect JSON data, and find effect's display name (based on the class name)
        raw_effects_list = json.loads(openshot.EffectInfo.Json())
        for raw_effect in raw_effects_list:
            if raw_effect.get("class_name") == self.effect_class:
                self.effect_name = raw_effect.get("name")
                break

        # Access C++ timeline and find the Clip instance which this effect should be applied to
        timeline_instance = get_app().window.timeline_sync.timeline
        for clip_instance in timeline_instance.Clips():
            if clip_instance.Id() == self.clip_id:
                self.clip_instance = clip_instance
                break

        # Load UI from designer & init
        ui_util.load_ui(self, self.ui_path)
        ui_util.init_ui(self)

        # get translations
        _ = get_app()._tr

        # Update window title
        self.setWindowTitle(self.windowTitle() % _(self.effect_name))

        # Pause playback
        get_app().window.PauseSignal.emit()

        # Track metrics
        track_metric_screen("process-effect-screen")

        # Loop through options and create widgets
        form_layout = self.scrollAreaWidgetContents.layout()
        for param in effect_params:
            # Create Label
            widget = None
            label = QLabel()
            label.setText(_(param["title"]))
            label.setToolTip(_(param["title"]))

            if param["type"] == "link":
                # create a clickable link
                label.setText('<a href="%s">%s</a>' % (param["value"], _(param["title"])))
                label.setTextInteractionFlags(Qt.TextBrowserInteraction)
                label.linkActivated.connect(functools.partial(self.link_activated, widget, param))

            if param["type"] in ("download-yolo", "download-yolo5", "download-object-mask"):
                widget = self.create_model_download_widget(param)

            if param["type"] == "spinner":
                # create QDoubleSpinBox
                widget = QDoubleSpinBox()
                widget.setMinimum(float(param["min"]))
                widget.setMaximum(float(param["max"]))
                widget.setValue(float(param["value"]))
                widget.setSingleStep(1.0)
                widget.setToolTip(_(param["title"]))
                widget.valueChanged.connect(functools.partial(self.spinner_value_changed, widget, param))

                # Set initial context
                self.context[param["setting"]] = float(param["value"])

            if param["type"] == "rect":
                # create QPushButton which opens up a display of the clip, with ability to select Rectangle
                widget = RegionButton(_("Click to Select"))
                widget.setMinimumHeight(80)
                widget.setToolTip(_(param["title"]))
                widget.clicked.connect(functools.partial(self.rect_select_clicked, widget, param))

                # Set initial context
                self.context[param["setting"]] = {"button-clicked": False, "x": 0, "y": 0, "width": 0, "height": 0}

            if param["type"] == "object-mask-selection":
                widget = QWidget()
                widget_layout = QHBoxLayout(widget)
                widget_layout.setContentsMargins(0, 0, 0, 0)
                widget_layout.setSpacing(6)

                select_button = QPushButton(_("Select Point / Region"))
                select_button.setToolTip(_("Choose a positive point or rectangle, and optional negative points."))
                select_button.clicked.connect(functools.partial(self.object_mask_select_clicked, select_button, param))

                status_widget = QLabel()
                status_widget.setMinimumWidth(20)
                status_widget.setToolTip(_("Point or region has not been selected."))

                select_button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
                status_widget.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
                widget_layout.addWidget(select_button, 1, Qt.AlignVCenter)
                widget_layout.addWidget(status_widget, 0, Qt.AlignVCenter)

                self.context[param["setting"]] = {}
                self.selection_fields[param["setting"]] = {
                    "param": param,
                    "button": select_button,
                    "status": status_widget,
                    "valid": False,
                }

            if param["type"] == "spinner-int":
                # create QDoubleSpinBox
                widget = QSpinBox()
                widget.setMinimum(int(param["min"]))
                widget.setMaximum(int(param["max"]))
                widget.setValue(int(param["value"]))
                widget.setSingleStep(1)
                widget.setToolTip(_(param["title"]))
                widget.valueChanged.connect(functools.partial(self.spinner_value_changed, widget, param))

                # Set initial context
                self.context[param["setting"]] = int(param["value"])

            elif param["type"] == "text":
                # create QLineEdit
                widget = QLineEdit()
                widget.setText(_(param["value"]))
                widget.textChanged.connect(functools.partial(self.text_value_changed, widget, param))

                # Set initial context
                self.context[param["setting"]] = param["value"]

            elif param["type"] == "file":
                widget = QWidget()
                widget_layout = QHBoxLayout(widget)
                widget_layout.setContentsMargins(0, 0, 0, 0)
                widget_layout.setSpacing(6)

                path_widget = QLineEdit()
                path_widget.setText(param["value"])
                path_widget.setToolTip(_(param["title"]))
                path_widget.textChanged.connect(functools.partial(self.file_value_changed, path_widget, param))

                status_widget = QLabel()
                status_widget.setMinimumWidth(20)
                status_widget.setToolTip(_("File has not been validated yet."))

                browse_button = QPushButton(_("Browse"))
                browse_button.setToolTip(_("Choose a file from disk."))
                browse_button.clicked.connect(functools.partial(self.file_browse_clicked, path_widget, param))

                path_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
                status_widget.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
                browse_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)

                widget_layout.addWidget(path_widget, 1, Qt.AlignVCenter)
                widget_layout.addWidget(status_widget, 0, Qt.AlignVCenter)
                widget_layout.addWidget(browse_button, 0, Qt.AlignVCenter)

                self.context[param["setting"]] = param["value"]
                self.file_fields[param["setting"]] = {
                    "param": param,
                    "path": path_widget,
                    "status": status_widget,
                }

            elif param["type"] == "bool":
                # create spinner
                widget = QCheckBox()
                if param["value"] == True:
                    widget.setCheckState(Qt.Checked)
                    self.context[param["setting"]] = True
                else:
                    widget.setCheckState(Qt.Unchecked)
                    self.context[param["setting"]] = False
                widget.stateChanged.connect(functools.partial(self.bool_value_changed, widget, param))

            elif param["type"] == "dropdown":

                # create spinner
                widget = QComboBox()

                # Get values
                value_list = param["values"]

                # Add normal values
                box_index = 0
                for value_item in value_list:
                    k = value_item["name"]
                    v = value_item["value"]
                    i = value_item.get("icon", None)

                    # add dropdown item
                    widget.addItem(_(k), v)

                    # select dropdown (if default)
                    if v == param["value"]:
                        widget.setCurrentIndex(box_index)

                        # Set initial context
                        self.context[param["setting"]] = param["value"]
                    box_index = box_index + 1

                widget.currentIndexChanged.connect(functools.partial(self.dropdown_index_changed, widget, param))

            # Add Label and Widget to the form
            if widget and label:
                # Add minimum size
                label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
                label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
                widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)

                form_layout.addRow(label, widget)
                if param.get("advanced"):
                    label.setVisible(self.advanced_visible)
                    widget.setVisible(self.advanced_visible)
                    self.advanced_rows.append((label, widget))
                if param["type"] in ("download-yolo", "download-yolo5", "download-object-mask"):
                    for group in self.download_groups:
                        if group["param"] is param:
                            status_label = QLabel(_("Status"))
                            status_label.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
                            status_label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
                            form_layout.addRow(status_label, group["status-widget"])
                            break

            elif not widget and label:
                label.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
                label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
                form_layout.addRow(label)

        # Add buttons
        self.cancel_button = QPushButton(_('Cancel'))
        self.process_button = QPushButton(_('Process Effect'))
        self.buttonBox.addButton(self.process_button, QDialogButtonBox.AcceptRole)
        self.buttonBox.addButton(self.cancel_button, QDialogButtonBox.RejectRole)
        self.sync_download_groups_to_selected_models()
        self.update_file_validation()

        # flag to close the clip processing thread
        self.cancel_clip_processing = False
        self.effect = None

    def load_manifest_models(self, key, path):
        """Load and cache a model manifest by resource path."""
        if key not in self.model_manifests:
            self.model_manifests[key] = load_model_manifest(path)
        return self.model_manifests[key]

    def create_model_download_widget(self, param):
        """Create the shared model dropdown, download button, status, and advanced toggle."""
        _ = get_app()._tr
        group_type = param["type"]
        if group_type in ("download-yolo", "download-yolo5"):
            self.load_yolo_models()
            models = self.yolo_models
            selected_model = recommended_model(models)
            button_tooltip = _("Download selected YOLO model files.")
            button_callback = self.download_yolo_clicked
        else:
            manifest = self.load_manifest_models("cutie", CUTIE_MODELS_PATH)
            models = manifest.get("models", [])
            selected_model = recommended_model(models)
            button_tooltip = _("Download selected Object Mask model files.")
            button_callback = self.download_object_mask_clicked

        widget = QWidget()
        control_layout = QHBoxLayout(widget)
        control_layout.setContentsMargins(0, 0, 0, 0)
        control_layout.setSpacing(6)

        model_combo = QComboBox()
        selected_index = 0
        selected_id = selected_model.get("id") if selected_model else None
        for index, model in enumerate(models):
            label_text = translated_model_label(model, YOLO_FALLBACK_MODEL_NAME)
            model_combo.addItem(label_text, model.get("id"))
            if model.get("id") == selected_id:
                selected_index = index

        download_button = QPushButton(model_download_button_label(selected_model))
        download_button.setToolTip(button_tooltip)
        download_button.clicked.connect(
            functools.partial(button_callback, download_button, model_combo, param)
        )

        if models:
            model_combo.setCurrentIndex(selected_index)
            model_combo.currentIndexChanged.connect(
                functools.partial(self.model_download_index_changed, model_combo, download_button, param)
            )

        model_combo.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        download_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        control_layout.addWidget(model_combo, 1, Qt.AlignVCenter)
        control_layout.addWidget(download_button, 0, Qt.AlignVCenter)

        status_widget = QWidget()
        status_layout = QHBoxLayout(status_widget)
        status_layout.setContentsMargins(0, 0, 0, 0)
        status_layout.setSpacing(6)

        status_label = QLabel(_("Not Downloaded"))
        status_label.setStyleSheet("color: #e53935; font-weight: bold;")
        status_label.setToolTip(_("Required model files are not downloaded."))

        advanced_button = QPushButton(_("Advanced"))
        advanced_button.setCheckable(True)
        advanced_button.setToolTip(_("Show model file paths."))
        advanced_button.setFlat(True)
        advanced_button.setObjectName("advancedLinkButton")
        advanced_button.setStyleSheet(
            "QPushButton#advancedLinkButton {"
            "border: none; background: transparent; padding: 0;"
            "text-decoration: underline; color: palette(link);"
            "}"
            "QPushButton#advancedLinkButton:hover { color: palette(highlight); }"
        )
        advanced_button.clicked.connect(functools.partial(self.advanced_clicked, advanced_button))

        status_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        advanced_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        status_layout.addWidget(status_label, 1, Qt.AlignVCenter)
        status_layout.addWidget(advanced_button, 0, Qt.AlignVCenter)

        self.download_groups.append({
            "param": param,
            "type": group_type,
            "combo": model_combo,
            "button": download_button,
            "status": status_label,
            "status-widget": status_widget,
            "file-settings": list(param.get("file-settings", [])),
        })
        self.advanced_buttons.append(advanced_button)
        return widget

    def advanced_clicked(self, button, checked):
        """Show or hide advanced file path rows."""
        _ = get_app()._tr
        self.advanced_visible = bool(checked)
        for label, widget in self.advanced_rows:
            label.setVisible(self.advanced_visible)
            widget.setVisible(self.advanced_visible)
        for advanced_button in self.advanced_buttons:
            advanced_button.blockSignals(True)
            advanced_button.setChecked(self.advanced_visible)
            advanced_button.setText(_("Hide Advanced") if self.advanced_visible else _("Advanced"))
            advanced_button.blockSignals(False)

    def sync_download_groups_to_selected_models(self):
        """Apply selected manifest entries to hidden file path fields."""
        for group in self.download_groups:
            self.apply_download_group_selection(group)

    def selected_download_group_model(self, group):
        """Return the selected model for a download group."""
        model_id = group["combo"].itemData(group["combo"].currentIndex())
        if group["type"] in ("download-yolo", "download-yolo5"):
            return self.yolo_model_by_id(model_id)
        manifest = self.load_manifest_models("cutie", CUTIE_MODELS_PATH)
        for model in manifest.get("models", []):
            if model.get("id") == model_id:
                return model
        return recommended_model(manifest.get("models", []))

    def apply_download_group_selection(self, group):
        """Update hidden file path fields for the selected model group."""
        model = self.selected_download_group_model(group)
        if not model:
            return
        group["button"].setText(model_download_button_label(model))
        if group["type"] in ("download-yolo", "download-yolo5"):
            self.set_yolo_file_fields(group["param"], model)
        else:
            self.set_object_mask_file_fields(group["param"], model)

    def set_file_field_path(self, setting, path):
        """Update a file field path and context without emitting intermediate callbacks."""
        if not setting or not path:
            return
        self.context[setting] = path
        field = self.file_fields.get(setting)
        if field:
            field["path"].blockSignals(True)
            field["path"].setText(path)
            field["path"].blockSignals(False)

    def model_download_index_changed(self, combo_widget, download_button, param, index):
        """Model download dropdown selection changed."""
        for group in self.download_groups:
            if group["combo"] is combo_widget:
                self.apply_download_group_selection(group)
                self.start_file_validation_wait_cursor(group)
                self.file_validation_timer.start()
                return

    def start_file_validation_wait_cursor(self, group=None):
        """Show the wait cursor while selected model file paths are being validated."""
        if group:
            group["ready"] = False
            group["validating"] = True
        model = self.selected_download_group_model(group) if group else None
        log.debug(
            "Validating selected model paths: type=%s id=%s name=%s paths=%s",
            group.get("type") if group else "",
            model.get("id") if model else "",
            model.get("name") if model else "",
            {
                setting: self.context.get(setting)
                for setting in (group.get("file-settings", []) if group else [])
            },
        )
        if not self.validation_wait_cursor:
            QApplication.setOverrideCursor(Qt.WaitCursor)
            self.validation_wait_cursor = True
        self.update_selection_validation()
        if hasattr(self, "process_button"):
            self.process_button.setEnabled(False)
        QCoreApplication.processEvents()

    def restore_file_validation_wait_cursor(self):
        """Restore the cursor after model file validation completes."""
        if self.validation_wait_cursor:
            QApplication.restoreOverrideCursor()
            self.validation_wait_cursor = False

    def link_activated(self, widget, param, value):
        """Link activated"""
        webbrowser.open(value, new=1)

    def model_message(self, message, model=None):
        """Format a translated message with the object detection model name."""
        model_name = (model or {}).get("name") or YOLO_FALLBACK_MODEL_NAME
        return message % {"model": model_name}

    def load_yolo_models(self):
        """Load packaged YOLO model metadata for the Object Detection initializer."""
        if self.yolo_models_manifest is None:
            self.yolo_models_manifest = load_yolo_models_manifest()
            self.yolo_models = self.yolo_models_manifest.get("models", [])

    def yolo_model_by_id(self, model_id):
        """Return a YOLO model manifest entry by id."""
        for model in self.yolo_models:
            if model.get("id") == model_id:
                return model
        return self.yolo_models[0] if self.yolo_models else None

    def selected_yolo_model(self, widget):
        """Return the model selected in the YOLO download dropdown."""
        return self.yolo_model_by_id(widget.itemData(widget.currentIndex()))

    def yolo_model_index_changed(self, widget, download_button, param, index):
        """YOLO model selection changed."""
        model = self.yolo_model_by_id(widget.itemData(index))
        if model:
            download_button.setText(yolo_download_button_label(model))
            self.set_yolo_file_fields(param, model)

    def efficient_sam_model_by_id(self, model_id=None):
        """Return an EfficientSAM manifest entry by id or the recommended entry."""
        manifest = self.load_manifest_models("efficient-sam", EFFICIENT_SAM_MODELS_PATH)
        models = manifest.get("models", [])
        for model in models:
            if model.get("id") == model_id:
                return model
        return recommended_model(models)

    def object_mask_cutie_paths(self, cutie_model):
        """Return Object Mask Cutie paths for the selected quality tier."""
        model_files = CUTIE_MODEL_FILES.get(cutie_model.get("id"), CUTIE_MODEL_FILES["cutie-medium"])
        install_dir = model_install_dir(cutie_model)
        return {key: os.path.join(install_dir, filename) for key, filename in model_files.items()}

    def object_mask_efficient_sam_path(self, efficient_sam_model):
        """Return the Object Mask EfficientSAM path for the selected SAM model."""
        filename = EFFICIENT_SAM_MODEL_FILES.get(
            efficient_sam_model.get("id"),
            EFFICIENT_SAM_MODEL_FILES["efficient-sam-tiny-1024"],
        )
        return os.path.join(model_install_dir(efficient_sam_model), filename)

    def set_object_mask_file_fields(self, param, cutie_model):
        """Point the Object Mask file inputs at the selected quality files."""
        efficient_sam_model = self.efficient_sam_model_by_id()
        self.context["cutie_model_dir"] = model_install_dir(cutie_model)
        self.context["model_size"] = 1024
        path_updates = {
            param.get("efficient-sam-setting"): self.object_mask_efficient_sam_path(efficient_sam_model),
        }
        cutie_paths = self.object_mask_cutie_paths(cutie_model)
        for file_key, setting in (param.get("cutie-settings") or {}).items():
            path_updates[setting] = cutie_paths.get(file_key)

        log.debug(
            "Object Mask model selection applied: cutie_id=%s cutie_name=%s efficient_sam_id=%s paths=%s",
            cutie_model.get("id") if cutie_model else "",
            cutie_model.get("name") if cutie_model else "",
            efficient_sam_model.get("id") if efficient_sam_model else "",
            path_updates,
        )
        for setting, path in path_updates.items():
            self.set_file_field_path(setting, path)

    def spinner_value_changed(self, widget, param, value):
        """Spinner value change callback"""
        self.context[param["setting"]] = value
        log.info(self.context)

    def bool_value_changed(self, widget, param, state):
        """Boolean value change callback"""
        if state == Qt.Checked:
            self.context[param["setting"]] = True
        else:
            self.context[param["setting"]] = False
        log.info(self.context)

    def dropdown_index_changed(self, widget, param, index):
        """Dropdown value change callback"""
        value = widget.itemData(index)
        self.context[param["setting"]] = value
        log.info(self.context)

    def text_value_changed(self, widget, param, value=None):
        """Textbox value change callback"""
        try:
            # Attempt to load value from QTextEdit (i.e. multi-line)
            if not value:
                value = widget.toPlainText()
        except:
            log.debug('Failed to get plain text from widget')

        self.context[param["setting"]] = value
        log.info(self.context)

    def file_value_changed(self, widget, param, value=None):
        """File path textbox change callback"""
        if value is None:
            value = widget.text()
        self.context[param["setting"]] = value
        log.debug("File path changed, scheduling validation: setting=%s path=%s", param["setting"], value)
        group = next(
            (
                download_group
                for download_group in self.download_groups
                if param["setting"] in download_group.get("file-settings", [])
            ),
            None,
        )
        self.start_file_validation_wait_cursor(group)
        self.file_validation_timer.start()
        log.info(self.context)

    def file_browse_clicked(self, widget, param):
        """Browse for a file path."""
        _ = get_app()._tr
        current_path = widget.text()
        start_dir = os.path.dirname(current_path) if current_path else info.USER_PATH
        selected_path = QFileDialog.getOpenFileName(
            self,
            _("Choose %s") % _(param["title"]),
            start_dir,
            _(param.get("file-filter", "All files (*)")),
        )[0]
        if selected_path:
            widget.setText(selected_path)

    def validate_onnx_model_load(self, path):
        """Return whether OpenCV can load this ONNX model."""
        _ = get_app()._tr
        try:
            stat = os.stat(path)
        except OSError:
            return False, _("File not found")

        cache_key = (os.path.abspath(path), stat.st_mtime_ns, stat.st_size)
        cached = self.onnx_validation_cache.get(cache_key)
        if cached:
            return cached

        try:
            error_text = openshot.ClipProcessingJobs.ValidateONNXModel(path)
        except Exception as ex:
            error_text = str(ex)

        if not error_text:
            result = (True, _("Ready"))
        else:
            log.error("ONNX model validation failed: %s", error_text)
            result = (False, error_text)
            if "Unsupported data type: FLOAT16" in error_text:
                result = (
                    False,
                    self.model_message(_("%(model)s requires an FP32 model file for this OpenCV build.")),
                )

        self.onnx_validation_cache[cache_key] = result
        return result

    def validate_file_param(self, param, path):
        """Validate a pre-processing file field."""
        _ = get_app()._tr
        if param.get("required") and not path:
            return False, _("Required file")
        if not path:
            return True, ""
        if not os.path.isfile(path):
            return False, _("File not found")

        validator = param.get("validator")
        if validator == "onnx":
            if not path.lower().endswith(".onnx"):
                return False, _("Expected a model file.")
            if os.path.getsize(path) <= 0:
                return False, _("Model file is empty.")
            return self.validate_onnx_model_load(path)
        if validator == "classes":
            try:
                with open(path, "r", encoding="utf-8") as classes_file:
                    class_names = [line.strip() for line in classes_file if line.strip()]
            except OSError:
                return False, _("Unable to read class names")
            if not class_names:
                return False, _("Class names file is empty")
        return True, _("Ready")

    @staticmethod
    def object_mask_payload_to_source_pixels(payload, preview_size, source_size):
        """Scale annotation payload coordinates from preview pixels to source-frame pixels."""
        if not isinstance(payload, dict):
            return payload
        if not preview_size or not source_size:
            return payload

        try:
            preview_width = float(preview_size.width())
            preview_height = float(preview_size.height())
            source_width = float(source_size.width())
            source_height = float(source_size.height())
        except Exception:
            return payload
        if preview_width <= 0.0 or preview_height <= 0.0 or source_width <= 0.0 or source_height <= 0.0:
            return payload

        scale_x = source_width / preview_width
        scale_y = source_height / preview_height
        scaled = json.loads(json.dumps(payload))

        for frame_payload in (scaled.get("frames") or {}).values():
            if not isinstance(frame_payload, dict):
                continue
            for key in ("positive_points", "negative_points"):
                for point in frame_payload.get(key) or []:
                    if not isinstance(point, dict):
                        continue
                    point["x"] = float(point.get("x", 0.0)) * scale_x
                    point["y"] = float(point.get("y", 0.0)) * scale_y
            for key in ("positive_rects", "negative_rects"):
                for rect in frame_payload.get(key) or []:
                    if not isinstance(rect, dict):
                        continue
                    rect["x1"] = float(rect.get("x1", 0.0)) * scale_x
                    rect["x2"] = float(rect.get("x2", 0.0)) * scale_x
                    rect["y1"] = float(rect.get("y1", 0.0)) * scale_y
                    rect["y2"] = float(rect.get("y2", 0.0)) * scale_y

        return scaled

    @staticmethod
    def object_mask_context_from_payload(payload):
        """Convert SelectRegion annotation payload into ObjectMask preprocessing settings."""
        if not isinstance(payload, dict):
            return {}, False
        frames = payload.get("frames") or {}
        if not isinstance(frames, dict) or not frames:
            return {}, False

        try:
            seed_frame_key = str(payload.get("seed_frame") or sorted(int(f) for f in frames.keys())[0])
        except Exception:
            seed_frame_key = sorted(frames.keys())[0]
        frame_payload = frames.get(seed_frame_key) or frames.get(str(seed_frame_key)) or {}

        context = {}
        positive_points = frame_payload.get("positive_points") or []
        negative_points = frame_payload.get("negative_points") or []
        positive_rects = frame_payload.get("positive_rects") or []
        negative_rects = frame_payload.get("negative_rects") or []
        context["positive_points"] = list(positive_points)
        context["negative_points"] = list(negative_points)
        context["positive_rects"] = list(positive_rects)
        context["negative_rects"] = list(negative_rects)

        if positive_rects:
            rect = positive_rects[0]
            x1 = float(rect.get("x1", 0.0))
            y1 = float(rect.get("y1", 0.0))
            x2 = float(rect.get("x2", 0.0))
            y2 = float(rect.get("y2", 0.0))
            context.update({
                "rect_x1": min(x1, x2),
                "rect_y1": min(y1, y2),
                "rect_x2": max(x1, x2),
                "rect_y2": max(y1, y2),
            })
            if not positive_points:
                context.update({
                    "positive_x": (x1 + x2) / 2.0,
                    "positive_y": (y1 + y2) / 2.0,
                })

        if positive_points:
            point = positive_points[0]
            context.update({
                "positive_x": float(point.get("x", 0.0)),
                "positive_y": float(point.get("y", 0.0)),
            })

        if negative_points:
            point = negative_points[0]
            context.update({
                "negative_x": float(point.get("x", 0.0)),
                "negative_y": float(point.get("y", 0.0)),
            })

        return context, "positive_x" in context and "positive_y" in context

    def object_mask_processing_frame_size(self, payload):
        """Return the frame size that libopenshot will use while processing ObjectMask."""
        if not isinstance(payload, dict):
            return QSize()
        try:
            frames = payload.get("frames") or {}
            seed_frame = int(payload.get("seed_frame") or sorted(int(f) for f in frames.keys())[0])
        except Exception:
            seed_frame = 1
        try:
            frame = self.clip_instance.GetFrame(max(1, seed_frame))
            if frame:
                width = int(frame.GetWidth())
                height = int(frame.GetHeight())
                if width > 0 and height > 0:
                    return QSize(width, height)
        except Exception as ex:
            log.warning("Unable to read Object Mask processing frame size: %s", ex)
        return QSize()

    def update_selection_validation(self):
        """Update selection indicators and return whether all required selections are valid."""
        _ = get_app()._tr
        all_valid = True
        object_mask_ready = all(
            group.get("ready")
            for group in self.download_groups
            if group.get("type") == "download-object-mask"
        )
        for setting, field in self.selection_fields.items():
            valid = bool(field.get("valid"))
            button = field.get("button")
            if button is not None:
                button.setEnabled(object_mask_ready)
                if object_mask_ready:
                    button.setToolTip(_("Choose a positive point or rectangle, and optional negative points."))
                else:
                    button.setToolTip(_("Download valid Object Mask model files before selecting points."))
            status = field["status"]
            status.setText("✓" if valid else "✕")
            status.setStyleSheet("color: #4caf50; font-weight: bold;" if valid else "color: #e53935; font-weight: bold;")
            if valid:
                status.setToolTip(_("Ready"))
            elif object_mask_ready:
                status.setToolTip(_("Select a positive point or rectangle."))
            else:
                status.setToolTip(_("Object Mask model files are not ready."))
            all_valid = all_valid and valid
        return all_valid

    def update_file_validation(self):
        """Update validation indicators and Process button state."""
        if self.file_validation_timer.isActive():
            self.file_validation_timer.stop()

        try:
            all_valid = True
            file_results = {}
            for setting, field in self.file_fields.items():
                path = field["path"].text()
                valid, message = self.validate_file_param(field["param"], path)
                file_results[setting] = {
                    "valid": valid,
                    "message": message,
                    "path": path,
                }
                status = field["status"]
                status.setText("✓" if valid else "✕")
                status.setStyleSheet("color: #4caf50; font-weight: bold;" if valid else "color: #e53935; font-weight: bold;")
                status.setToolTip(message)
                all_valid = all_valid and valid

            self.update_download_group_statuses(file_results)
            all_valid = all_valid and self.update_selection_validation()

            if hasattr(self, "process_button"):
                self.process_button.setEnabled(all_valid and not self.processing_effect)
        finally:
            self.restore_file_validation_wait_cursor()

    def update_download_group_statuses(self, file_results):
        """Update visible status text for model download groups."""
        _ = get_app()._tr
        for group in self.download_groups:
            expected_settings = group.get("file-settings", [])
            results = [
                file_results[setting]
                for setting in expected_settings
                if setting in file_results
            ]
            status = group["status"]
            group["validating"] = False
            partial_downloads = self.download_group_partial_files(group)
            if not results or len(results) != len(expected_settings):
                group["ready"] = False
                log.debug(
                    "Model path validation incomplete: type=%s expected=%s available=%s",
                    group.get("type"),
                    expected_settings,
                    sorted(file_results.keys()),
                )
                status.setText(_("Not Downloaded"))
                status.setStyleSheet("color: #e53935; font-weight: bold;")
                status.setToolTip(_("Required model files are not downloaded."))
                continue

            missing = any(
                not result.get("path") or result.get("message") in (_("File not found"), _("Required file"))
                for result in results
            )
            valid = all(result.get("valid") for result in results)
            group["ready"] = valid and not partial_downloads
            log.debug(
                "Model path validation complete: type=%s ready=%s valid=%s partials=%s results=%s",
                group.get("type"),
                group.get("ready"),
                valid,
                partial_downloads,
                {
                    setting: {
                        "valid": file_results[setting].get("valid"),
                        "path": file_results[setting].get("path"),
                        "message": file_results[setting].get("message"),
                    }
                    for setting in expected_settings
                    if setting in file_results
                },
            )
            if valid and not partial_downloads:
                status.setText(_("Ready"))
                status.setStyleSheet("color: #4caf50; font-weight: bold;")
                status.setToolTip(_("All required model files are ready."))
            elif partial_downloads:
                status.setText(_("Not Downloaded"))
                status.setStyleSheet("color: #e53935; font-weight: bold;")
                status.setToolTip(_("A model download is incomplete. Download again to finish installation."))
            elif missing:
                status.setText(_("Not Downloaded"))
                status.setStyleSheet("color: #e53935; font-weight: bold;")
                status.setToolTip(_("Required model files are not downloaded."))
            else:
                status.setText(_("Invalid"))
                status.setStyleSheet("color: #e53935; font-weight: bold;")
                status.setToolTip(_("One or more model files are invalid."))

    def download_group_partial_files(self, group):
        """Return leftover partial downloads for a model download group."""
        dirs = []
        if group.get("type") in ("download-yolo", "download-yolo5"):
            model = self.selected_download_group_model(group)
            if model:
                dirs.append(yolo_model_dir(model))
        elif group.get("type") == "download-object-mask":
            cutie_model = self.selected_download_group_model(group)
            efficient_sam_model = self.efficient_sam_model_by_id()
            if efficient_sam_model:
                dirs.append(model_install_dir(efficient_sam_model))
            if cutie_model:
                dirs.append(model_install_dir(cutie_model))

        partials = []
        for directory in dirs:
            try:
                for filename in os.listdir(directory):
                    if filename.endswith((".download", ".zip")):
                        partials.append(os.path.join(directory, filename))
            except OSError:
                continue
        return partials

    def set_processing_controls_enabled(self, enabled):
        """Enable or disable editable processing controls."""
        self.scrollArea.setEnabled(enabled)
        self.process_button.setEnabled(enabled)
        self.cancel_button.setEnabled(True)

    @staticmethod
    def cancel_processing_job(processing, max_wait_seconds=1.0):
        """Cancel a native processing job and tolerate its handle becoming invalid."""
        try:
            processing.CancelProcessing()
        except RuntimeError:
            # Cancellation can invalidate the SWIG-backed job immediately.
            return

        wait_start = time.time()
        while (time.time() - wait_start) < max_wait_seconds:
            try:
                if processing.IsDone():
                    return
            except RuntimeError:
                # An invalid handle after CancelProcessing means there is
                # nothing left for the dialog to poll.
                return
            QCoreApplication.processEvents()
            time.sleep(0.01)

    def file_sha256(self, path):
        """Return the SHA256 checksum of a file."""
        digest = hashlib.sha256()
        with open(path, "rb") as input_file:
            for chunk in iter(lambda: input_file.read(1024 * 1024), b""):
                digest.update(chunk)
        return digest.hexdigest()

    def yolo_installed_files_match(self, model):
        """Return whether an installed YOLO model matches recorded metadata."""
        model_path = yolo_model_path(model)
        classes_path = yolo_classes_path(model)
        metadata_path = os.path.join(yolo_model_dir(model), YOLO_INSTALL_METADATA)
        try:
            with open(metadata_path, "r", encoding="utf-8") as metadata_file:
                metadata = json.load(metadata_file)
            return (
                metadata.get("id") == model.get("id")
                and metadata.get("asset_sha256") == model.get("sha256")
                and os.path.isfile(model_path)
                and os.path.isfile(classes_path)
                and metadata.get("model_sha256") == self.file_sha256(model_path)
                and metadata.get("classes_sha256") == self.file_sha256(classes_path)
            )
        except (OSError, ValueError):
            return False

    def set_yolo_file_fields(self, param, model):
        """Point the YOLO file inputs at the selected model files."""
        for setting, path in (
            (param.get("model-setting"), yolo_model_path(model)),
            (param.get("classes-setting"), yolo_classes_path(model)),
        ):
            self.set_file_field_path(setting, path)

    def yolo_model_download_url(self, model):
        """Return the download URL for a YOLO model manifest entry."""
        base_url = self.yolo_models_manifest.get("base_url", "")
        return urljoin("%s/" % base_url.rstrip("/"), model.get("asset", ""))

    def model_download_url(self, manifest, model):
        """Return the download URL for a generic model manifest entry."""
        base_url = manifest.get("base_url", "")
        return urljoin("%s/" % base_url.rstrip("/"), model.get("asset", ""))

    def extract_yolo_zip_member(self, yolo_zip, suffixes, destination_path):
        """Extract the first matching file from a verified YOLO model archive."""
        if isinstance(suffixes, str):
            suffixes = (suffixes,)
        for member in yolo_zip.infolist():
            if member.is_dir():
                continue
            if os.path.basename(member.filename).lower().endswith(tuple(suffixes)):
                with yolo_zip.open(member) as source_file, open(destination_path, "wb") as output_file:
                    output_file.write(source_file.read())
                return
        raise ValueError("Downloaded YOLO files are invalid.")

    def extract_zip_members_to_dir(self, zip_path, destination_dir):
        """Extract all regular files from a verified model archive."""
        os.makedirs(destination_dir, exist_ok=True)
        with zipfile.ZipFile(zip_path) as model_zip:
            for member in model_zip.infolist():
                if member.is_dir():
                    continue
                filename = os.path.basename(member.filename)
                if not filename:
                    continue
                destination_path = os.path.join(destination_dir, filename)
                download_path = "{}.download".format(destination_path)
                with model_zip.open(member) as source_file, open(download_path, "wb") as output_file:
                    output_file.write(source_file.read())
                if os.path.getsize(download_path) <= 0:
                    os.remove(download_path)
                    raise ValueError("Downloaded model files are invalid.")
                os.replace(download_path, destination_path)

    def write_model_install_metadata(self, install_dir, metadata_name, model, installed_paths):
        """Record extracted-file hashes for future already-downloaded checks."""
        metadata_path = os.path.join(install_dir, metadata_name)
        metadata_download_path = "{}.download".format(metadata_path)
        with open(metadata_download_path, "w", encoding="utf-8") as metadata_file:
            json.dump(
                {
                    "id": model.get("id"),
                    "asset": model.get("asset"),
                    "asset_sha256": model.get("sha256"),
                    "files": {
                        os.path.basename(path): self.file_sha256(path)
                        for path in installed_paths
                    },
                },
                metadata_file,
                indent=2,
                sort_keys=True,
            )
            metadata_file.write("\n")
        os.replace(metadata_download_path, metadata_path)

    def installed_model_files_match(self, install_dir, metadata_name, model, installed_paths):
        """Return whether installed model files match recorded metadata."""
        metadata_path = os.path.join(install_dir, metadata_name)
        try:
            with open(metadata_path, "r", encoding="utf-8") as metadata_file:
                metadata = json.load(metadata_file)
            file_hashes = metadata.get("files", {})
            return (
                metadata.get("id") == model.get("id")
                and metadata.get("asset_sha256") == model.get("sha256")
                and all(
                    os.path.isfile(path)
                    and file_hashes.get(os.path.basename(path)) == self.file_sha256(path)
                    for path in installed_paths
                )
            )
        except (OSError, ValueError):
            return False

    def download_manifest_archive(self, manifest, model, install_dir, metadata_name, installed_paths, progress, start, end):
        """Download, verify, and extract a generic model archive."""
        _ = get_app()._tr
        os.makedirs(install_dir, exist_ok=True)
        if self.installed_model_files_match(install_dir, metadata_name, model, installed_paths):
            return

        zip_download_path = os.path.join(install_dir, "%s.download" % model.get("asset", "model.zip"))

        def report_progress(downloaded_size, total_size):
            if total_size > 0:
                fraction = min(1.0, downloaded_size / float(total_size))
                progress.setValue(int(start + (end - start) * fraction))
            else:
                progress.setValue(start)
            QCoreApplication.processEvents()
            if progress.wasCanceled():
                raise DownloadCancelled()

        try:
            http_client.download_file(
                self.model_download_url(manifest, model),
                zip_download_path,
                model_label(model),
                report_progress,
                cancel_exceptions=(DownloadCancelled,),
            )
            if self.file_sha256(zip_download_path) != model.get("sha256"):
                raise ValueError(_("Downloaded model files are invalid."))
            self.extract_zip_members_to_dir(zip_download_path, install_dir)
            if not all(os.path.isfile(path) and os.path.getsize(path) > 0 for path in installed_paths):
                raise ValueError(_("Downloaded model files are invalid."))
            self.write_model_install_metadata(install_dir, metadata_name, model, installed_paths)
        finally:
            if os.path.exists(zip_download_path):
                os.remove(zip_download_path)

    def write_yolo_install_metadata(self, model, model_path, classes_path):
        """Record extracted-file hashes for future already-downloaded checks."""
        metadata_path = os.path.join(yolo_model_dir(model), YOLO_INSTALL_METADATA)
        metadata_download_path = "{}.download".format(metadata_path)
        with open(metadata_download_path, "w", encoding="utf-8") as metadata_file:
            json.dump(
                {
                    "id": model.get("id"),
                    "asset": model.get("asset"),
                    "asset_sha256": model.get("sha256"),
                    "model": YOLO_MODEL_FILENAME,
                    "model_sha256": self.file_sha256(model_path),
                    "classes": YOLO_CLASSES_FILENAME,
                    "classes_sha256": self.file_sha256(classes_path),
                },
                metadata_file,
                indent=2,
                sort_keys=True,
            )
            metadata_file.write("\n")
        os.replace(metadata_download_path, metadata_path)

    def download_yolo_clicked(self, widget, combo_widget, param):
        """Download the selected YOLO model files."""
        _ = get_app()._tr
        model = self.selected_yolo_model(combo_widget)
        if not model:
            QMessageBox.warning(self, _("Download Failed"), _("No YOLO models are available."))
            return

        model_dir = yolo_model_dir(model)
        model_path = yolo_model_path(model)
        classes_path = yolo_classes_path(model)
        os.makedirs(model_dir, exist_ok=True)

        if self.yolo_installed_files_match(model):
            self.set_yolo_file_fields(param, model)
            self.update_file_validation()
            QMessageBox.information(
                self,
                self.model_message(_("Download %(model)s Files"), model),
                self.model_message(_("The %(model)s files are already downloaded."), model),
            )
            return

        progress = QProgressDialog(
            self.model_message(_("Downloading %(model)s files..."), model),
            _("Cancel"),
            0,
            100,
            self,
        )
        progress.setWindowTitle(self.model_message(_("Download %(model)s Files"), model))
        progress.setWindowModality(Qt.WindowModal)
        progress.setMinimumDuration(0)
        zip_download_path = os.path.join(model_dir, "%s.download" % model.get("asset", "model.zip"))
        model_download_path = "{}.download".format(model_path)
        classes_download_path = "{}.download".format(classes_path)
        metadata_download_path = os.path.join(model_dir, "%s.download" % YOLO_INSTALL_METADATA)

        def remove_partial_downloads():
            for path in (zip_download_path, model_download_path, classes_download_path, metadata_download_path):
                if os.path.exists(path):
                    os.remove(path)

        def report_progress(downloaded_size, total_size):
            if total_size > 0:
                progress.setValue(min(100, int(downloaded_size * 100 / total_size)))
            else:
                progress.setValue(0)
            QCoreApplication.processEvents()
            if progress.wasCanceled():
                raise DownloadCancelled()

        try:
            http_client.download_file(
                self.yolo_model_download_url(model),
                zip_download_path,
                self.model_message(_("%(model)s files"), model),
                report_progress,
                cancel_exceptions=(DownloadCancelled,),
            )
            if self.file_sha256(zip_download_path) != model.get("sha256"):
                raise ValueError(self.model_message(_("Downloaded %(model)s files are invalid."), model))

            with zipfile.ZipFile(zip_download_path) as yolo_zip:
                self.extract_yolo_zip_member(yolo_zip, ".onnx", model_download_path)
                self.extract_yolo_zip_member(yolo_zip, (".names", ".txt"), classes_download_path)

            if os.path.getsize(model_download_path) <= 0 or os.path.getsize(classes_download_path) <= 0:
                raise ValueError(self.model_message(_("Downloaded %(model)s files are invalid."), model))

            os.replace(model_download_path, model_path)
            os.replace(classes_download_path, classes_path)
            self.write_yolo_install_metadata(model, model_path, classes_path)
        except DownloadCancelled:
            remove_partial_downloads()
            log.info("YOLO file download cancelled")
            self.update_file_validation()
            return
        except Exception as ex:
            remove_partial_downloads()
            progress.close()
            QMessageBox.warning(self, _("Download Failed"), str(ex))
            log.error("Failed to download YOLO files: %s", ex)
            self.update_file_validation()
            return
        finally:
            if os.path.exists(zip_download_path):
                os.remove(zip_download_path)
            progress.close()

        self.set_yolo_file_fields(param, model)
        self.update_file_validation()

    def download_object_mask_clicked(self, widget, combo_widget, param):
        """Download the selected Object Mask model files."""
        _ = get_app()._tr
        cutie_manifest = self.load_manifest_models("cutie", CUTIE_MODELS_PATH)
        cutie_model = self.selected_download_group_model({
            "type": "download-object-mask",
            "combo": combo_widget,
        })
        efficient_sam_manifest = self.load_manifest_models("efficient-sam", EFFICIENT_SAM_MODELS_PATH)
        efficient_sam_model = self.efficient_sam_model_by_id()
        if not cutie_model or not efficient_sam_model:
            QMessageBox.warning(self, _("Download Failed"), _("No Object Mask models are available."))
            return

        self.set_object_mask_file_fields(param, cutie_model)
        cutie_paths_by_key = self.object_mask_cutie_paths(cutie_model)
        cutie_paths = [
            cutie_paths_by_key[file_key]
            for file_key in ("encode-key", "encode-value", "memory-readout", "decode")
        ]
        efficient_sam_path = self.object_mask_efficient_sam_path(efficient_sam_model)
        efficient_sam_install_dir = model_install_dir(efficient_sam_model)
        cutie_install_dir = model_install_dir(cutie_model)

        efficient_sam_ready = self.installed_model_files_match(
            efficient_sam_install_dir,
            EFFICIENT_SAM_INSTALL_METADATA,
            efficient_sam_model,
            [efficient_sam_path],
        )
        cutie_ready = self.installed_model_files_match(
            cutie_install_dir,
            CUTIE_INSTALL_METADATA,
            cutie_model,
            cutie_paths,
        )

        if efficient_sam_ready and cutie_ready:
            self.update_file_validation()
            QMessageBox.information(
                self,
                _("Download Object Mask Files"),
                _("The Object Mask files are already downloaded."),
            )
            return

        progress = QProgressDialog(
            _("Downloading Object Mask files..."),
            _("Cancel"),
            0,
            100,
            self,
        )
        progress.setWindowTitle(_("Download Object Mask Files"))
        progress.setWindowModality(Qt.WindowModal)
        progress.setMinimumDuration(0)
        progress.setValue(0)

        if efficient_sam_ready:
            efficient_sam_range = None
            cutie_range = (0, 100)
        elif cutie_ready:
            efficient_sam_range = (0, 100)
            cutie_range = None
        else:
            efficient_sam_range = (0, 35)
            cutie_range = (35, 100)

        try:
            if efficient_sam_range:
                self.download_manifest_archive(
                    efficient_sam_manifest,
                    efficient_sam_model,
                    efficient_sam_install_dir,
                    EFFICIENT_SAM_INSTALL_METADATA,
                    [efficient_sam_path],
                    progress,
                    efficient_sam_range[0],
                    efficient_sam_range[1],
                )
            if cutie_range:
                self.download_manifest_archive(
                    cutie_manifest,
                    cutie_model,
                    cutie_install_dir,
                    CUTIE_INSTALL_METADATA,
                    cutie_paths,
                    progress,
                    cutie_range[0],
                    cutie_range[1],
                )
        except DownloadCancelled:
            log.info("Object Mask file download cancelled")
            self.update_file_validation()
            return
        except Exception as ex:
            QMessageBox.warning(self, _("Download Failed"), str(ex))
            log.error("Failed to download Object Mask files: %s", ex)
            self.update_file_validation()
            return
        finally:
            progress.close()

        self.set_object_mask_file_fields(param, cutie_model)
        self.update_file_validation()

    def rect_select_clicked(self, widget, param):
        """Rect select button clicked"""
        _ = get_app()._tr
        self.context[param["setting"]].update({"button-clicked": True})

        # show dialog
        from windows.region import SelectRegion
        from classes.query import File, Clip

        c = Clip.get(id=self.clip_id)
        reader_path = c.data.get('reader', {}).get('path','')
        f = File.get(path=reader_path)
        if f:
            win = SelectRegion(f, self.clip_instance, parent=self)
            # Run the dialog event loop - blocking interaction on this window during that time
            result = win.exec_()
            if result == QDialog.Accepted:
                # self.first_frame = win.current_frame
                # Region selected (get coordinates if any)
                selected_rect = win.selected_rect_normalized() if hasattr(win, "selected_rect_normalized") else None
                if selected_rect:
                    x1 = float(selected_rect.get("normalized_x", 0.0))
                    y1 = float(selected_rect.get("normalized_y", 0.0))
                    xw = float(selected_rect.get("normalized_width", 0.0))
                    yh = float(selected_rect.get("normalized_height", 0.0))
                else:
                    topLeft = win.videoPreview.regionTopLeftHandle
                    bottomRight = win.videoPreview.regionBottomRightHandle
                    curr_frame_size = win.videoPreview.curr_frame_size
                    if not topLeft or not bottomRight or not curr_frame_size:
                        QMessageBox.warning(
                            self,
                            _("Invalid Region"),
                            _("Please draw a rectangle region before clicking Select Region."),
                        )
                        return
                    x1 = topLeft.x() / curr_frame_size.width()
                    y1 = topLeft.y() / curr_frame_size.height()
                    x2 = bottomRight.x() / curr_frame_size.width()
                    y2 = bottomRight.y() / curr_frame_size.height()
                    xw = x2 - x1
                    yh = y2 - y1

                # Get QImage of region
                region_qimage = win.selected_region_qimage() if hasattr(win, "selected_region_qimage") else None
                if region_qimage is None and win.videoPreview.region_qimage:
                    region_qimage = win.videoPreview.region_qimage
                if region_qimage:

                    # Resize QImage to match button size
                    resized_qimage = region_qimage.scaled(widget.size(), Qt.IgnoreAspectRatio, Qt.SmoothTransformation)

                    # Remove button text (so region QImage is more visible)
                    widget.setImage(resized_qimage)
                    widget.setText("")

                # If data found, add to context
                self.context[param["setting"]].update({"normalized_x": x1, "normalized_y": y1,
                                                       "normalized_width": xw,
                                                       "normalized_height": yh,
                                                       "first-frame": win.current_frame,
                                                       })
                log.info(self.context)

        else:
            log.error('No file found with path: %s' % reader_path)

    def object_mask_select_clicked(self, widget, param):
        """Open annotation selector for Object Mask prompts."""
        _ = get_app()._tr
        from windows.region import SelectRegion
        from classes.query import File, Clip

        self.sync_download_groups_to_selected_models()
        self.update_file_validation()
        object_mask_groups = [
            group for group in self.download_groups
            if group.get("type") == "download-object-mask"
        ]
        if object_mask_groups and not all(group.get("ready") for group in object_mask_groups):
            QMessageBox.warning(
                self,
                _("Object Mask Models Not Ready"),
                _("Download valid Object Mask model files before selecting points."),
            )
            return

        c = Clip.get(id=self.clip_id)
        reader_path = c.data.get('reader', {}).get('path', '')
        f = File.get(path=reader_path)
        if not f:
            log.error('No file found with path: %s' % reader_path)
            return

        log.debug(
            "Opening Object Mask selector with preview models: efficient_sam=%s cutie_key=%s "
            "cutie_value=%s cutie_memory=%s cutie_decode=%s processing_device=%s",
            self.context.get("efficient_sam_model"),
            self.context.get("cutie_encode_key_model"),
            self.context.get("cutie_encode_value_model"),
            self.context.get("cutie_memory_readout_model"),
            self.context.get("cutie_decode_model"),
            self.context.get("processing-device") or self.context.get("processing_device"),
        )
        win = SelectRegion(
            f,
            self.clip_instance,
            selection_mode="annotate",
            parent=self,
            object_mask_preview_context=dict(self.context),
        )
        result = win.exec_()
        if result != QDialog.Accepted:
            return

        payload = win.selection_payload() if hasattr(win, "selection_payload") else {}
        preview_size = getattr(win.videoPreview, "curr_frame_size", None)
        processing_size = self.object_mask_processing_frame_size(payload)
        if not processing_size.isValid() or processing_size.width() <= 0 or processing_size.height() <= 0:
            processing_size = QSize(int(getattr(win, "width", 0) or 0), int(getattr(win, "height", 0) or 0))
        log.debug(
            "Object Mask selection scale: preview=%sx%s processing=%sx%s",
            preview_size.width() if preview_size else 0,
            preview_size.height() if preview_size else 0,
            processing_size.width(),
            processing_size.height(),
        )
        payload = self.object_mask_payload_to_source_pixels(payload, preview_size, processing_size)

        prompt_context, valid = self.object_mask_context_from_payload(payload)
        if not valid:
            QMessageBox.warning(
                self,
                _("Invalid Selection"),
                _("Please select a positive point or rectangle."),
            )
            return

        self.context[param["setting"]] = payload
        self.context.update(prompt_context)

        field = self.selection_fields.get(param["setting"])
        if field:
            field["valid"] = True
            includes = len(prompt_context.get("positive_points") or []) + len(prompt_context.get("positive_rects") or [])
            excludes = len(prompt_context.get("negative_points") or []) + len(prompt_context.get("negative_rects") or [])
            if includes == 0 and "positive_x" in prompt_context:
                includes = 1
            label_args = {
                "include": includes,
                "exclude": excludes,
            }
            if excludes > 0:
                button_text = _("Selection: %(include)s include, %(exclude)s exclude") % label_args
            else:
                button_text = _("Selection: %(include)s include") % label_args
            field["button"].setText(
                button_text
            )
        self.update_file_validation()
        log.info(self.context)

    def accept(self):
        """ Start processing effect """
        _ = get_app()._tr
        self.update_file_validation()
        if self.file_fields and not self.process_button.isEnabled():
            return

        # Enable ProgressBar
        self.progressBar.setEnabled(True)
        self.processing_effect = True
        self.set_processing_controls_enabled(False)

        # Print effect settings
        log.info(self.context)

        # Create effect Id and protobuf data path
        ID = get_app().project.generate_id()

        # Create protobuf data path
        protobufPath = os.path.join(info.PROTOBUF_DATA_PATH, ID + '.data')
        if os.name == 'nt' : protobufPath = protobufPath.replace("\\", "/")

        self.context["protobuf_data_path"] = protobufPath

        # Load into JSON string info about protobuf data path
        jsonString = json.dumps(self.context)

        def show_processing_error(message=None):
            """Show a processing failure and leave the dialog open for correction."""
            if not message:
                message = _("Unable to process this effect. Check the model files and settings.")
            self.progressBar.setEnabled(False)
            self.processing_effect = False
            self.set_processing_controls_enabled(True)
            self.update_file_validation()
            QMessageBox.warning(self, _("Processing Failed"), message)

        # Generate processed data
        try:
            processing = openshot.ClipProcessingJobs(self.effect_class, jsonString)
            processing.processClip(self.clip_instance, jsonString)
        except Exception as ex:
            log.error("Failed to start effect processing: %s", ex)
            show_processing_error(str(ex))
            return

        # get processing status
        blank_error_start = None
        while(not processing.IsDone() ):
            if processing.GetError():
                message = (processing.GetErrorMessage() or "").strip()
                if message:
                    log.error("Effect processing failed: %s", message)
                    self.cancel_processing_job(processing)
                    show_processing_error(message)
                    return
                if blank_error_start is None:
                    blank_error_start = time.time()
                elif time.time() - blank_error_start > 3.0:
                    log.error("Effect processing failed without an error message")
                    self.cancel_processing_job(processing)
                    show_processing_error()
                    return
            else:
                blank_error_start = None

            # update progressbar
            progressionStatus = processing.GetProgress()
            self.progressBar.setValue(int(progressionStatus))
            time.sleep(0.01)

            # Process any queued events
            QCoreApplication.processEvents()

            # if the cancel button was pressed, close the processing thread
            if(self.cancel_clip_processing):
                self.cancel_processing_job(processing)
                return

        if not self.cancel_clip_processing and processing.GetError():
            message = (processing.GetErrorMessage() or "").strip()
            log.error(
                "Effect processing failed%s",
                ": %s" % message if message else "",
            )
            show_processing_error(message)
            return

        if(not self.cancel_clip_processing):
            # Load processed data into effect
            self.effect = openshot.EffectInfo().CreateEffect(self.effect_class)
            self.effect.SetJson( '{"protobuf_data_path": "%s"}' % protobufPath )
            self.effect.Id(ID)

            # Accept dialog
            self.restore_file_validation_wait_cursor()
            super(ProcessEffect, self).accept()

    def reject(self):
        # Cancel dialog
        self.exporting = False
        self.cancel_clip_processing = True
        self.restore_file_validation_wait_cursor()
        super(ProcessEffect, self).reject()



"""
 @file
 @brief Unit tests for OpenShot release details helpers
 @author OpenShot Studios, LLC

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.
 """

import os
import sys
import unittest


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes import release_details


class ReleaseDetailsTests(unittest.TestCase):
    def test_release_details_url_accepts_official_versions(self):
        self.assertEqual(
            release_details.release_details_url("3.5.1"),
            "https://www.openshot.org/releases/3.5.1/",
        )

    def test_release_details_url_skips_development_versions(self):
        self.assertIsNone(release_details.release_details_url("3.5.1-dev"))

    def test_release_details_url_skips_release_candidates(self):
        self.assertIsNone(release_details.release_details_url("3.5.1-rc1"))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import json
import os
import threading
import time
from pathlib import Path

from qt_api import QEvent, QObject, QTimer
from qt_api import QAction, QDialog, QDockWidget

from classes import info
from classes.app import get_app
from classes.logger import log


def _env_true(name, default=False):
    value = os.getenv(name)
    if value is None:
        return default
    return str(value).strip().lower() in ("1", "true", "yes", "on")


def _env_float(name, default):
    value = os.getenv(name)
    if value is None:
        return float(default)
    try:
        return float(value)
    except Exception:
        return float(default)


class UiTraceRecorder(QObject):
    """Record UI traces in a centralized, env-configured subsystem."""

    def __init__(self, window):
        super().__init__(window)
        self.window = window
        self._lock = threading.Lock()
        self._seq = 0
        self._closed = False
        self._updates_handle = None
        self._selections_handle = None
        self._events_handle = None
        self._dialog_connections = {}
        self._action_connections = []
        self._dock_connections = []
        self._cache_timer = None
        self._last_cache_signature = None
        self._updates_listener_connected = False
        self._selection_connected = False
        self._dialog_filter_installed = False

        self._include_load = _env_true("OPENSHOT_UI_TRACE_INCLUDE_LOAD", default=False)
        self._include_ignored = _env_true("OPENSHOT_UI_TRACE_INCLUDE_IGNORED", default=False)
        self._trace_actions = _env_true("OPENSHOT_UI_TRACE_ACTIONS", default=True)
        self._trace_dialogs = _env_true("OPENSHOT_UI_TRACE_DIALOGS", default=True)
        self._trace_docks = _env_true("OPENSHOT_UI_TRACE_DOCKS", default=True)
        self._trace_cache = _env_true("OPENSHOT_UI_TRACE_CACHE", default=False)
        self._cache_interval_ms = max(100, int(_env_float("OPENSHOT_UI_TRACE_CACHE_INTERVAL", 0.5) * 1000))
        self.enabled = False

        updates_path, selections_path, events_path = self._resolve_paths()
        if not updates_path and not selections_path and not events_path:
            return

        try:
            if updates_path:
                self._updates_handle = self._open_jsonl(updates_path)
            if selections_path:
                self._selections_handle = self._open_jsonl(selections_path)
            if events_path:
                self._events_handle = self._open_jsonl(events_path)

            if self._updates_handle or self._events_handle:
                get_app().updates.add_listener(self)
                self._updates_listener_connected = True
            if self._selections_handle or self._events_handle:
                self.window.SelectionChanged.connect(self._on_selection_changed)
                self._selection_connected = True
            if self._events_handle and self._trace_actions:
                self._connect_actions()
            if self._events_handle and self._trace_dialogs:
                get_app().installEventFilter(self)
                self._dialog_filter_installed = True
            if self._events_handle and self._trace_docks:
                self._connect_docks()
            if self._events_handle and self._trace_cache:
                self._start_cache_timer()

            self.enabled = True
            self._write_meta()
            log.info(
                "UI trace recorder enabled. updates=%s selections=%s events=%s",
                updates_path,
                selections_path,
                events_path,
            )
        except Exception:
            log.error("Failed to initialize UI trace recorder", exc_info=1)
            self.close()

    def _resolve_paths(self):
        updates_path = os.getenv("OPENSHOT_UI_TRACE_UPDATES", "").strip()
        selections_path = os.getenv("OPENSHOT_UI_TRACE_SELECTIONS", "").strip()
        events_path = os.getenv("OPENSHOT_UI_TRACE_EVENTS", "").strip()

        trace_dir = os.getenv("OPENSHOT_UI_TRACE_DIR", "").strip()
        trace_enabled = _env_true("OPENSHOT_UI_TRACE", default=False)

        if trace_dir and (trace_enabled or updates_path or selections_path or events_path):
            stamp = time.strftime("%Y%m%d_%H%M%S")
            pid = os.getpid()
            trace_root = Path(trace_dir).expanduser().resolve()
            trace_root.mkdir(parents=True, exist_ok=True)
            if not updates_path:
                updates_path = str(trace_root / f"updates_{stamp}_{pid}.jsonl")
            if not selections_path:
                selections_path = str(trace_root / f"selections_{stamp}_{pid}.jsonl")
            if not events_path:
                events_path = str(trace_root / f"events_{stamp}_{pid}.jsonl")

        if not trace_enabled and not updates_path and not selections_path and not events_path:
            return None, None, None

        return updates_path or None, selections_path or None, events_path or None

    @staticmethod
    def _open_jsonl(path):
        p = Path(path).expanduser().resolve()
        p.parent.mkdir(parents=True, exist_ok=True)
        return open(p, "a", encoding="utf-8", buffering=1)

    def _next_seq(self):
        with self._lock:
            self._seq += 1
            return self._seq

    def _event_base(self, kind):
        return {
            "event": kind,
            "seq": self._next_seq(),
            "ts": round(time.time(), 6),
        }

    def _write_jsonl(self, handle, payload):
        if handle is None or self._closed:
            return
        line = json.dumps(payload, ensure_ascii=False, default=str)
        handle.write(line + "\n")

    def _write_event(self, payload, include_updates=False, include_selections=False):
        self._write_jsonl(self._events_handle, payload)
        if include_updates:
            self._write_jsonl(self._updates_handle, payload)
        if include_selections:
            self._write_jsonl(self._selections_handle, payload)

    def _write_meta(self):
        payload = self._event_base("meta")
        payload.update(
            {
                "pid": os.getpid(),
                "include_load": self._include_load,
                "include_ignored": self._include_ignored,
                "trace_actions": self._trace_actions,
                "trace_dialogs": self._trace_dialogs,
                "trace_docks": self._trace_docks,
                "trace_cache": self._trace_cache,
                "cache_interval_ms": self._cache_interval_ms,
                "project_path": str(get_app().project.current_filepath or ""),
            }
        )
        self._write_event(payload, include_updates=True, include_selections=True)

    def changed(self, action):
        if (self._updates_handle is None and self._events_handle is None) or self._closed:
            return
        try:
            ignore_history = bool(get_app().updates.ignore_history)
            if action.type == "load" and not self._include_load:
                return
            if ignore_history and not self._include_ignored:
                return
            payload = self._event_base("update")
            payload.update(
                {
                    "action_type": action.type,
                    "transaction": action.transaction,
                    "key": action.key,
                    "value": action.values,
                    "old_values": action.old_values,
                    "ignore_history": ignore_history,
                    "data_version": int(get_app().updates.data_version),
                }
            )
            self._write_event(payload, include_updates=True)
        except Exception:
            log.error("UI trace update write failed", exc_info=1)

    def _on_selection_changed(self):
        if (self._selections_handle is None and self._events_handle is None) or self._closed:
            return
        try:
            payload = self._event_base("selection")
            payload.update(
                {
                    "selected_items": json.loads(json.dumps(self.window.selected_items)),
                    "selected_clips": list(self.window.selected_clips),
                    "selected_transitions": list(self.window.selected_transitions),
                    "selected_effects": list(self.window.selected_effects),
                    "selected_tracks": list(self.window.selected_tracks),
                    "selected_markers": list(self.window.selected_markers),
                    "show_property_id": self.window.show_property_id,
                    "show_property_type": self.window.show_property_type,
                }
            )
            self._write_event(payload, include_selections=True)
        except Exception:
            log.error("UI trace selection write failed", exc_info=1)

    def _connect_actions(self):
        for action in self.window.findChildren(QAction):
            try:
                slot = self._make_action_slot(action)
                action.triggered.connect(slot)
                self._action_connections.append((action, slot))
            except Exception:
                log.debug("Failed to connect QAction trace for %s", action.objectName(), exc_info=1)

    def _make_action_slot(self, action):
        def _slot(checked=False):
            self._on_action_triggered(action, checked)

        return _slot

    def _on_action_triggered(self, action, checked):
        if self._events_handle is None or self._closed:
            return
        payload = self._event_base("action_triggered")
        payload.update(
            {
                "action_name": action.objectName() or "",
                "action_text": (action.text() or "").replace("&", ""),
                "checked": bool(checked),
            }
        )
        self._write_event(payload)

    def _connect_docks(self):
        for dock in self.window.findChildren(QDockWidget):
            try:
                visibility_slot = self._make_dock_visibility_slot(dock)
                dock.visibilityChanged.connect(visibility_slot)
                self._dock_connections.append((dock.visibilityChanged, visibility_slot))
            except Exception:
                log.debug("Failed to connect dock visibility trace for %s", dock.objectName(), exc_info=1)

    def _make_dock_visibility_slot(self, dock):
        def _slot(visible):
            self._on_dock_visibility_changed(dock, visible)

        return _slot

    def _on_dock_visibility_changed(self, dock, visible):
        if self._events_handle is None or self._closed:
            return
        payload = self._event_base("dock_visibility")
        payload.update(
            {
                "object_name": dock.objectName() or "",
                "window_title": dock.windowTitle() or "",
                "visible": bool(visible),
                "floating": bool(dock.isFloating()),
            }
        )
        self._write_event(payload)

    def eventFilter(self, obj, event):
        if self._closed or not self._trace_dialogs or self._events_handle is None:
            return False
        try:
            if isinstance(obj, QDialog):
                event_type = event.type()
                if event_type == QEvent.Show:
                    self._attach_dialog(obj)
                    self._emit_dialog("shown", obj)
                elif event_type == QEvent.Hide:
                    self._emit_dialog("hidden", obj)
                elif event_type == QEvent.Close:
                    self._emit_dialog("closed", obj)
        except Exception:
            log.debug("UI trace dialog eventFilter failed", exc_info=1)
        return False

    def _attach_dialog(self, dialog):
        key = id(dialog)
        if key in self._dialog_connections:
            return
        slots = []
        try:
            accepted_slot = lambda: self._emit_dialog("accepted", dialog, result="accepted")
            dialog.accepted.connect(accepted_slot)
            slots.append((dialog.accepted, accepted_slot))
        except Exception:
            pass
        try:
            rejected_slot = lambda: self._emit_dialog("rejected", dialog, result="rejected")
            dialog.rejected.connect(rejected_slot)
            slots.append((dialog.rejected, rejected_slot))
        except Exception:
            pass
        try:
            finished_slot = lambda code: self._emit_dialog("finished", dialog, result=int(code))
            dialog.finished.connect(finished_slot)
            slots.append((dialog.finished, finished_slot))
        except Exception:
            pass
        self._dialog_connections[key] = slots

    def _emit_dialog(self, phase, dialog, result=None):
        payload = self._event_base("dialog_lifecycle")
        payload.update(
            {
                "phase": phase,
                "class_name": dialog.__class__.__name__,
                "object_name": dialog.objectName() or "",
                "window_title": dialog.windowTitle() or "",
                "modal": bool(dialog.isModal()),
            }
        )
        if result is not None:
            payload["result"] = result
        self._write_event(payload)

    def _start_cache_timer(self):
        self._cache_timer = QTimer(self)
        self._cache_timer.setInterval(self._cache_interval_ms)
        self._cache_timer.timeout.connect(self._sample_cache_progress)
        self._cache_timer.start()

    @staticmethod
    def _metric_call(obj, name):
        attr = getattr(obj, name, None)
        if attr is None:
            return None
        try:
            return attr() if callable(attr) else attr
        except Exception:
            return None

    def _collect_cache_state(self):
        state = {
            "current_frame": int(getattr(self.window, "current_frame", 0) or 0),
        }
        cache_obj = getattr(self.window, "cache_object", None)
        if cache_obj is not None:
            state["cache_class"] = cache_obj.__class__.__name__
            metrics = {}
            for name in (
                "Count",
                "GetCount",
                "Size",
                "GetSize",
                "Bytes",
                "GetBytes",
                "FrameCount",
                "GetFrameCount",
                "MaxBytes",
                "GetMaxBytes",
            ):
                value = self._metric_call(cache_obj, name)
                if value is not None:
                    metrics[name] = value
            if metrics:
                state["metrics"] = metrics

        preview_cache = Path(info.PREVIEW_CACHE_PATH)
        if preview_cache.exists():
            file_count = 0
            total_bytes = 0
            for child in preview_cache.iterdir():
                try:
                    if child.is_file():
                        file_count += 1
                        total_bytes += child.stat().st_size
                except Exception:
                    continue
            state["preview_cache_files"] = file_count
            state["preview_cache_bytes"] = total_bytes
        return state

    def _sample_cache_progress(self):
        if self._events_handle is None or self._closed:
            return
        try:
            state = self._collect_cache_state()
            signature = json.dumps(state, sort_keys=True, default=str)
            if signature == self._last_cache_signature:
                return
            self._last_cache_signature = signature
            payload = self._event_base("cache_progress")
            payload.update(state)
            self._write_event(payload)
        except Exception:
            log.debug("UI trace cache sampler failed", exc_info=1)

    def close(self):
        if self._closed:
            return
        self._closed = True

        if self._cache_timer:
            try:
                self._cache_timer.stop()
            except Exception:
                pass
            self._cache_timer = None

        try:
            if self._updates_listener_connected:
                get_app().updates.disconnect_listener(self)
        except Exception:
            pass
        self._updates_listener_connected = False

        try:
            if self._selection_connected:
                self.window.SelectionChanged.disconnect(self._on_selection_changed)
        except Exception:
            pass
        self._selection_connected = False

        if self._dialog_filter_installed:
            try:
                get_app().removeEventFilter(self)
            except Exception:
                pass
            self._dialog_filter_installed = False

        for action, slot in self._action_connections:
            try:
                action.triggered.disconnect(slot)
            except Exception:
                pass
        self._action_connections = []

        for signal, slot in self._dock_connections:
            try:
                signal.disconnect(slot)
            except Exception:
                pass
        self._dock_connections = []

        for slots in self._dialog_connections.values():
            for signal, slot in slots:
                try:
                    signal.disconnect(slot)
                except Exception:
                    pass
        self._dialog_connections.clear()

        for handle in (self._updates_handle, self._selections_handle, self._events_handle):
            try:
                if handle:
                    handle.flush()
                    handle.close()
            except Exception:
                pass
        self._updates_handle = None
        self._selections_handle = None
        self._events_handle = None



"""
 @file
 @brief Camera motion framing helpers for clip motion presets
 @author OpenShot Studios

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 """

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Optional, Tuple


PAN_AUTO = "auto"
PAN_LEFT_TO_RIGHT = "left_to_right"
PAN_RIGHT_TO_LEFT = "right_to_left"
PAN_TOP_TO_BOTTOM = "top_to_bottom"
PAN_BOTTOM_TO_TOP = "bottom_to_top"
PAN_LEFT = "pan_left"
PAN_RIGHT = "pan_right"
PAN_UP = "pan_up"
PAN_DOWN = "pan_down"

KEN_BURNS_AUTO = "auto"
KEN_BURNS_LEFT_TO_RIGHT = "left_to_right"
KEN_BURNS_RIGHT_TO_LEFT = "right_to_left"
KEN_BURNS_TOP_TO_BOTTOM = "top_to_bottom"
KEN_BURNS_BOTTOM_TO_TOP = "bottom_to_top"


@dataclass(frozen=True)
class CameraKeyframes:
    """Plain keyframe values returned by camera motion helpers."""

    scale_x: Tuple[float, float]
    scale_y: Tuple[float, float]
    location_x: Tuple[float, float] = (0.0, 0.0)
    location_y: Tuple[float, float] = (0.0, 0.0)


def _positive_float(value, fallback: float) -> float:
    try:
        value = float(value)
    except (TypeError, ValueError):
        return fallback
    return value if value > 0.0 else fallback


def _crop_base_size(
    project_width,
    project_height,
    source_width,
    source_height,
) -> Tuple[float, float]:
    """Return source size after SCALE_CROP with scale_x/scale_y at 1.0."""
    pw = _positive_float(project_width, 1920.0)
    ph = _positive_float(project_height, 1080.0)
    sw = _positive_float(source_width, pw)
    sh = _positive_float(source_height, ph)

    project_aspect = pw / ph
    source_aspect = sw / sh
    if source_aspect >= project_aspect:
        return ph * source_aspect, ph
    return pw, pw / source_aspect


def _safe_location(canvas_size: float, base_size: float, scale: float) -> float:
    """Largest normalized location value that keeps SCALE_CROP edges covered."""
    scaled_size = max(0.0001, float(base_size) * max(0.001, float(scale)))
    canvas_size = max(0.0001, float(canvas_size))
    if scaled_size <= canvas_size:
        return 0.0
    return (scaled_size - canvas_size) / (scaled_size + canvas_size)


def _scale_for_safe_location(canvas_size: float, base_size: float, target: float) -> float:
    """Return minimum scale needed for target normalized pan room on one axis."""
    target = max(0.0, min(float(target), 0.9))
    canvas_size = max(0.0001, float(canvas_size))
    base_size = max(0.0001, float(base_size))
    return (canvas_size * (1.0 + target)) / (base_size * (1.0 - target))


def _axis_for_direction(direction: str) -> str:
    if direction in (
            PAN_LEFT, PAN_RIGHT, PAN_LEFT_TO_RIGHT, PAN_RIGHT_TO_LEFT,
            KEN_BURNS_LEFT_TO_RIGHT, KEN_BURNS_RIGHT_TO_LEFT):
        return "x"
    if direction in (
            PAN_UP, PAN_DOWN, PAN_TOP_TO_BOTTOM, PAN_BOTTOM_TO_TOP,
            KEN_BURNS_TOP_TO_BOTTOM, KEN_BURNS_BOTTOM_TO_TOP):
        return "y"
    return "auto"


def _auto_ken_burns_direction(
    project_width,
    project_height,
    source_width,
    source_height,
) -> str:
    """Choose the strongest natural crop axis for a Ken Burns drift."""
    pw = _positive_float(project_width, 1920.0)
    ph = _positive_float(project_height, 1080.0)
    bw, bh = _crop_base_size(pw, ph, source_width, source_height)
    x_room = _safe_location(pw, bw, 1.0)
    y_room = _safe_location(ph, bh, 1.0)

    if x_room > y_room + 0.03:
        return KEN_BURNS_LEFT_TO_RIGHT
    if y_room > x_room + 0.03:
        return KEN_BURNS_BOTTOM_TO_TOP
    return KEN_BURNS_LEFT_TO_RIGHT


def _pan_endpoints(direction: str, magnitude: float) -> Tuple[float, float]:
    magnitude = round(max(0.0, float(magnitude)), 6)
    if direction in (PAN_LEFT, PAN_RIGHT_TO_LEFT, KEN_BURNS_RIGHT_TO_LEFT):
        return -magnitude, magnitude
    if direction in (PAN_RIGHT, PAN_LEFT_TO_RIGHT, KEN_BURNS_LEFT_TO_RIGHT):
        return magnitude, -magnitude
    if direction in (PAN_UP, PAN_BOTTOM_TO_TOP, KEN_BURNS_BOTTOM_TO_TOP):
        return -magnitude, magnitude
    if direction in (PAN_DOWN, PAN_TOP_TO_BOTTOM, KEN_BURNS_TOP_TO_BOTTOM):
        return magnitude, -magnitude
    return 0.0, 0.0


def camera_pan_keyframes(
    direction: str,
    project_width,
    project_height,
    source_width,
    source_height,
    *,
    target_pan: float = 0.18,
    edge_margin: float = 0.995,
) -> CameraKeyframes:
    """Return smart SCALE_CROP pan values for a requested camera pan direction."""
    pw = _positive_float(project_width, 1920.0)
    ph = _positive_float(project_height, 1080.0)
    if direction == PAN_AUTO:
        direction = _auto_ken_burns_direction(pw, ph, source_width, source_height)
    bw, bh = _crop_base_size(pw, ph, source_width, source_height)
    axis = _axis_for_direction(direction)
    canvas = pw if axis == "x" else ph
    base = bw if axis == "x" else bh

    natural_room = _safe_location(canvas, base, 1.0)
    if natural_room >= target_pan:
        scale = 1.0
        magnitude = natural_room * edge_margin
    else:
        scale = max(1.0, _scale_for_safe_location(canvas, base, target_pan))
        magnitude = _safe_location(canvas, base, scale) * edge_margin

    start, end = _pan_endpoints(direction, magnitude)
    if axis == "x":
        return CameraKeyframes((scale, scale), (scale, scale), (start, end), (0.0, 0.0))
    return CameraKeyframes((scale, scale), (scale, scale), (0.0, 0.0), (start, end))


def push_pull_keyframes(zoom_in: bool, *, zoom: float = 1.2) -> CameraKeyframes:
    """Return centered push-in or pull-out camera zoom keyframes."""
    start, end = (1.0, zoom) if zoom_in else (zoom, 1.0)
    return CameraKeyframes((start, end), (start, end))


def ken_burns_keyframes(
    zoom_in: bool,
    direction: str,
    project_width,
    project_height,
    source_width,
    source_height,
    *,
    zoom: float = 1.22,
    target_pan: float = 0.10,
    max_pan: float = 0.24,
) -> CameraKeyframes:
    """Return smart Ken Burns zoom and drift values."""
    pw = _positive_float(project_width, 1920.0)
    ph = _positive_float(project_height, 1080.0)
    bw, bh = _crop_base_size(pw, ph, source_width, source_height)

    if direction == KEN_BURNS_AUTO:
        direction = _auto_ken_burns_direction(pw, ph, source_width, source_height)

    axis = _axis_for_direction(direction)
    canvas = pw if axis == "x" else ph
    base = bw if axis == "x" else bh
    zoom = max(float(zoom), _scale_for_safe_location(canvas, base, target_pan))

    start_scale, end_scale = (1.0, zoom) if zoom_in else (zoom, 1.0)
    start_safe = min(_safe_location(canvas, base, start_scale) * 0.82, max_pan)
    end_safe = min(_safe_location(canvas, base, end_scale) * 0.82, max_pan)
    start_magnitude = max(start_safe, target_pan if start_safe else 0.0)
    end_magnitude = max(end_safe, target_pan if end_safe else 0.0)
    start_mag = _pan_endpoints(direction, min(start_magnitude, max_pan))[0]
    end_mag = _pan_endpoints(direction, min(end_magnitude, max_pan))[1]

    if axis == "x":
        return CameraKeyframes((start_scale, end_scale), (start_scale, end_scale), (start_mag, end_mag), (0.0, 0.0))
    return CameraKeyframes((start_scale, end_scale), (start_scale, end_scale), (0.0, 0.0), (start_mag, end_mag))


def source_dimensions_from_reader(reader: Optional[Dict]) -> Tuple[Optional[float], Optional[float]]:
    """Extract media dimensions from reader metadata."""
    if not isinstance(reader, dict):
        return None, None
    width = reader.get("width") or reader.get("display_width")
    height = reader.get("height") or reader.get("display_height")
    try:
        return float(width), float(height)
    except (TypeError, ValueError):
        return None, None



"""
 @file
 @brief This file contains the project file treeview, used by the main window
 @author Noah Figg <eggmunkee@hotmail.com>
 @author Jonathan Thomas <jonathan@openshot.org>
 @author Olivier Girard <eolinwen@gmail.com>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import uuid

from qt_api import QSize, Qt, QPoint
from qt_api import clear_override_cursor
from qt_api import modifiers_has
from qt_api import QDrag, QCursor, QPixmap, QPainter, QIcon, QColor, QFontMetrics
from qt_api import (
    QTreeView, QAbstractItemView, QSizePolicy, QHeaderView, QItemSelectionModel,
    QStyledItemDelegate, QStyleOptionViewItem, QStyle,
)

from classes import info
from classes.app import get_app
from classes.logger import log
from classes.query import File
from classes.qt_types import font_metrics_horizontal_advance
from .ai_tools_menu import add_ai_tools_menu
from .files_thumbnail_overlay import paint_media_overlay, paint_proxy_badge
from .menu import StyledContextMenu, add_bound_action
from .name_search_treeview import NameColumnKeyboardSearchMixin
from .optimized_preview_menu import add_optimized_preview_menu


def _is_generation_placeholder(file_id):
    return str(file_id or "").startswith("__genjob__:")


def _job_id_from_placeholder(file_id):
    file_id = str(file_id or "")
    if not _is_generation_placeholder(file_id):
        return None
    return file_id.split(":", 1)[1]


class FilesTreeProgressDelegate(QStyledItemDelegate):
    """Paint a thin progress line over thumbnail cells."""

    def __init__(self, view):
        super().__init__(view)
        self.view = view

    def paint(self, painter, option, index):
        super().paint(painter, option, index)

        if index.column() != 0:
            return

        opt = QStyleOptionViewItem(option)
        self.initStyleOption(opt, index)
        style = opt.widget.style() if opt.widget else self.view.style()
        deco_rect = style.subElementRect(QStyle.SE_ItemViewItemDecoration, opt, opt.widget)
        if not deco_rect.isValid():
            return

        media_type = index.sibling(index.row(), 3).data(Qt.DisplayRole)
        paint_media_overlay(painter, deco_rect, media_type)

        file_id = index.sibling(index.row(), 5).data(Qt.DisplayRole)
        queue = getattr(self.view.win, "generation_queue", None)
        generation_badge = None
        if file_id and queue:
            generation_badge = queue.get_file_badge(file_id)
            if not generation_badge and _is_generation_placeholder(file_id):
                job = queue.get_job(_job_id_from_placeholder(file_id))
                if job and job.get("status") in ("queued", "running", "canceling"):
                    label = "Queued" if job.get("status") == "queued" else "Generating"
                    generation_badge = {
                        "status": job.get("status"),
                        "progress": int(job.get("progress", 0)),
                        "label": label,
                        "job_id": job.get("id"),
                    }

        proxy_badge = None
        proxy_service = getattr(self.view.win, "proxy_service", None)
        if file_id and proxy_service and not _is_generation_placeholder(file_id):
            proxy_badge = proxy_service.get_file_badge(file_id)
            file_obj = File.get(id=file_id)
            if file_obj:
                paint_proxy_badge(painter, deco_rect, proxy_service.get_proxy_state(file_obj))

        self._paint_progress_bar(painter, deco_rect, generation_badge, QColor("#53A0ED"), 0)
        self._paint_progress_bar(painter, deco_rect, proxy_badge, QColor("#3AA1FF"), 1)

    def _paint_progress_bar(self, painter, deco_rect, badge, fill_color, stack_index):
        if not badge:
            return

        progress = int(badge.get("progress", 0))
        status = str(badge.get("status", "")).strip().lower()
        if status in ("queued", "running", "canceling"):
            progress = max(progress, 2)
        if progress <= 0:
            return

        bar_height = 3
        bar_margin = 2
        bottom_offset = (bar_height + 1) * int(stack_index)
        full_rect = deco_rect.adjusted(1, 0, -1, 0)
        full_rect.setTop(deco_rect.bottom() - bar_height - bar_margin - bottom_offset + 1)
        full_rect.setHeight(bar_height)
        if full_rect.width() <= 2:
            return

        fill_width = max(1, int((full_rect.width() * min(progress, 100)) / 100.0))
        fill_rect = full_rect.adjusted(0, 0, -(full_rect.width() - fill_width), 0)

        painter.save()
        painter.setPen(Qt.NoPen)
        painter.setBrush(QColor("#283241"))
        painter.drawRect(full_rect)
        painter.setBrush(fill_color)
        painter.drawRect(fill_rect)
        if status == "queued":
            label = "Queued"
            fm = QFontMetrics(painter.font())
            text_w = font_metrics_horizontal_advance(fm, label)
            text_h = fm.height()
            pad_x = 5
            pad_y = 2
            badge_w = text_w + (pad_x * 2)
            badge_h = text_h + (pad_y * 2)
            badge_bottom = full_rect.top() - 3
            badge_top = max(deco_rect.top() + 3, badge_bottom - badge_h + 1)
            badge_rect = deco_rect.adjusted(3, badge_top - deco_rect.top(), 0, 0)
            badge_rect.setWidth(badge_w)
            badge_rect.setHeight(badge_h)
            painter.setBrush(QColor(18, 22, 30, 220))
            painter.drawRoundedRect(badge_rect, 4, 4)
            painter.setPen(QColor("#EAF5FF"))
            painter.drawText(badge_rect, Qt.AlignCenter, label)
        painter.restore()


class FilesTreeView(NameColumnKeyboardSearchMixin, QTreeView):
    """ A TreeView QWidget used on the main window """
    drag_item_size = QSize(48, 48)
    drag_item_center = QPoint(24, 24)

    def contextMenuEvent(self, event):

        # Set context menu mode
        app = get_app()
        self.win = app.window
        _ = app._tr
        app.context_menu_object = "files"

        event.accept()
        index = self.indexAt(event.pos())
        if not index.isValid():
            self.clearSelection()
        else:
            self.selectionModel().setCurrentIndex(index, QItemSelectionModel.NoUpdate)

        # Build menu
        menu = StyledContextMenu(parent=self)

        add_bound_action(menu, self.win, "actionImportFiles", _("Import Files..."), "actionImportFiles_trigger")

        source_file = None
        active_job = None
        file_id = None
        if index.isValid():
            id_index = index.sibling(index.row(), 5)
            file_id = index.model().data(id_index, Qt.DisplayRole)
            if _is_generation_placeholder(file_id):
                job_id = _job_id_from_placeholder(file_id)
                queue = getattr(self.win, "generation_queue", None)
                active_job = queue.get_job(job_id) if queue else None
                if active_job and active_job.get("status") not in ("queued", "running", "canceling"):
                    active_job = None
            else:
                active_job = self.win.active_generation_job_for_file(file_id)
                source_file = File.get(id=file_id)
        add_ai_tools_menu(self.win, menu, source_file=source_file)

        if not active_job:
            self.win.actionGenerate.setEnabled(self.win.can_open_generate_dialog())
        if active_job:
            cancel_action = menu.addAction(_("Cancel Job"))
            delete_icon_path = os.path.join(info.PATH, "themes", "cosmic", "images", "track-delete-enabled.svg")
            if os.path.exists(delete_icon_path):
                cancel_action.setIcon(QIcon(delete_icon_path))
            else:
                cancel_action.setIcon(self.win.actionRemove_from_Project.icon())
            cancel_action.triggered.connect(
                lambda checked=False, job_id=active_job.get("id"): self.win.cancel_generation_job(job_id)
            )
        menu.addSeparator()
        add_bound_action(menu, self.win, "actionThumbnailView", _("Thumbnail View"), "actionThumbnailView_trigger")

        if index.isValid():
            # Look up the model item and our unique ID
            model = index.model()

            # Look up file_id from 5th column of row
            id_index = index.sibling(index.row(), 5)
            file_id = model.data(id_index, Qt.DisplayRole)

            # If a valid file selected, show file related options
            menu.addSeparator()

            # Add edit title option (if svg file)
            file = File.get(id=file_id)
            if not file:
                menu.popup(event.globalPos())
                return
            if file and file.data.get("path").endswith(".svg"):
                add_bound_action(menu, self.win, "actionEditTitle", _("Edit Title"), "actionEditTitle_trigger")
                add_bound_action(menu, self.win, "actionDuplicate", _("Duplicate"), "actionDuplicate_trigger")
                menu.addSeparator()

            add_bound_action(menu, self.win, "actionPreview_File", _("Preview File"), "actionPreview_File_trigger")
            add_optimized_preview_menu(self.win, menu)
            menu.addSeparator()
            add_bound_action(menu, self.win, "actionSplitFile", _("Split Clip"), "actionSplitFile_trigger")
            add_bound_action(menu, self.win, "actionExportFiles", _("Export Selected Clips"), "actionExportFiles_trigger")
            menu.addSeparator()
            add_bound_action(menu, self.win, "actionAdd_to_Timeline", _("Add to Timeline"), "actionAdd_to_Timeline_trigger")

            # Add Profile menu
            profile_menu = StyledContextMenu(title=_("Choose Profile"), parent=self)
            profile_icon = get_app().window.actionProfile.icon()
            profile_missing_icon = QIcon(":/icons/Humanity/actions/16/list-add.svg")
            profile_menu.setIcon(profile_icon)

            # Get file's profile
            file_profile = file.profile()
            if file_profile.info.description:
                action = profile_menu.addAction(profile_icon, file_profile.info.description)
                action.triggered.connect(lambda: get_app().window.actionProfile_trigger(file_profile))
            else:
                action = profile_menu.addAction(profile_missing_icon, _(f"Create Profile") + f": {file_profile.ShortName()}")
                action.triggered.connect(lambda: get_app().window.actionProfileEdit_trigger(file_profile))
            menu.addMenu(profile_menu)

            add_bound_action(menu, self.win, "actionFile_Properties", _("File Properties"), "actionFile_Properties_trigger")
            menu.addSeparator()
            add_bound_action(menu, self.win, "actionRemove_from_Project", _("Remove from Project"), "actionRemove_from_Project_trigger")
            menu.addSeparator()

        # Show menu
        menu.show_at(event)

    def mousePressEvent(self, event):
        index = self.indexAt(event.pos())
        if not index.isValid() and event.button() in (Qt.LeftButton, Qt.RightButton):
            self.clearSelection()
        super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event):
        # Get the index of the item at the click position
        index = self.indexAt(event.pos())
        if index.column() == 0:
            # If column 0 (thumbnail) is double-clicked, trigger the custom actions
            if modifiers_has(get_app().keyboardModifiers(), Qt.ShiftModifier):
                get_app().window.actionSplitFile.trigger()
            elif modifiers_has(get_app().keyboardModifiers(), Qt.ControlModifier):
                get_app().window.actionFile_Properties.trigger()
            else:
                get_app().window.actionPreview_File.trigger()
        else:
            super(FilesTreeView, self).mouseDoubleClickEvent(event)

    def dragEnterEvent(self, event):
        # If dragging urls onto widget, accept
        if event.mimeData().hasUrls():
            event.setDropAction(Qt.CopyAction)
            event.accept()

    def startDrag(self, supportedActions):
        """ Override startDrag method to display custom icon """

        # Get first column indexes for all selected rows
        selected = self.selectionModel().selectedRows(0)
        selected = [idx for idx in selected if not _is_generation_placeholder(idx.sibling(idx.row(), 5).data(Qt.DisplayRole))]

        # Check if there are any selected items
        if not selected:
            log.warning("No draggable items found in model!")
            return False

        # Get icons from up to 3 selected items
        icons = []
        for i in range(min(3, len(selected))):
            current = selected[i]
            icon = current.sibling(current.row(), 0).data(Qt.DecorationRole)
            if icon:
                icons.append(icon.pixmap(self.drag_item_size))

        # If no icons were retrieved, abort the drag
        if not icons:
            log.warning("No valid icons found for dragging!")
            return False

        # Calculate the total width of the composite pixmap including gaps
        gap = 1  # 1 pixel gap between icons
        total_width = (self.drag_item_size.width() * len(icons)) + (gap * (len(icons) - 1))

        # Create a composite pixmap to hold the icons in a row
        composite_pixmap = QPixmap(total_width, self.drag_item_size.height())
        composite_pixmap.fill(Qt.transparent)  # Start with a transparent background

        # Use a QPainter to draw the icons in a row with 1 pixel gap between them
        painter = QPainter(composite_pixmap)
        for idx, icon_pixmap in enumerate(icons):
            x_offset = idx * (self.drag_item_size.width() + gap)  # Position each icon with a gap
            painter.drawPixmap(int(x_offset), 0, icon_pixmap)
        painter.end()

        # Start the drag operation
        drag = QDrag(self)

        # Combine all selected items into the mime data
        mime_data = self.model().mimeData(selected)
        drag.setMimeData(mime_data)

        # Set the composite pixmap for the drag operation
        drag.setPixmap(composite_pixmap)

        # Set the hot spot to the center of the composite pixmap
        drag.setHotSpot(composite_pixmap.rect().center())

        # Start a transaction so all clips are grouped for a single undo
        tid = str(uuid.uuid4())
        get_app().updates.transaction_id = tid

        # Execute the drag operation
        exec_fn = getattr(drag, "exec", None) or getattr(drag, "exec_", None)
        if exec_fn is None:
            raise AttributeError("QDrag has no exec_/exec method")
        exec_fn(supportedActions)
        clear_override_cursor()

        # End transaction
        get_app().updates.transaction_id = None

    # Without defining this method, the 'copy' action doesn't show with cursor
    def dragMoveEvent(self, event):
        event.accept()

    # Handle a drag and drop being dropped on widget
    def dropEvent(self, event):
        if not event.mimeData().hasUrls():
            # Nothing we're interested in
            event.ignore()
            return
        event.accept()
        # Use try/finally so we always reset the cursor
        try:
            # Set cursor to waiting
            get_app().setOverrideCursor(QCursor(Qt.WaitCursor))

            qurl_list = event.mimeData().urls()
            log.info("Processing drop event for {} urls".format(len(qurl_list)))
            self.files_model.process_urls(qurl_list)
        finally:
            # Restore cursor
            get_app().restoreOverrideCursor()

    # Forward file-add requests to the model, for legacy code (previous API)
    def add_file(self, filepath):
        self.files_model.add_files(filepath)

    def filter_changed(self):
        self.refresh_view()

    def refresh_view(self):
        """Resize and hide certain columns"""
        self.hideColumn(3)
        self.hideColumn(4)
        self.hideColumn(5)
        self.resize_contents()

    def resize_contents(self):
        # Get size of widget
        thumbnail_width = 80
        tags_width = 75

        # Resize thumbnail and tags column
        self.header().resizeSection(0, thumbnail_width)
        self.header().resizeSection(2, tags_width)

        # Set stretch mode on certain columns
        self.header().setStretchLastSection(False)
        self.header().setSectionResizeMode(1, QHeaderView.Stretch)
        self.header().setSectionResizeMode(2, QHeaderView.Interactive)

    def value_updated(self, item):
        """ Name or tags updated """
        if self.files_model.ignore_updates:
            return
        if item is None:
            return
        if item.column() not in (1, 2):
            return

        # Determine what was changed
        file_id_item = self.files_model.model.item(item.row(), 5)
        name_item = self.files_model.model.item(item.row(), 1)
        tags_item = self.files_model.model.item(item.row(), 2)
        if not file_id_item or not name_item or not tags_item:
            return

        file_id = file_id_item.text()
        if _is_generation_placeholder(file_id):
            return

        name = name_item.text()
        tags = tags_item.text()

        # Get file object and update friendly name and tags attribute
        f = File.get(id=file_id)
        if not f:
            return
        f.data.update({"name": name or os.path.basename(f.data.get("path"))})
        if "tags" in f.data or tags:
            f.data.update({"tags": tags})

        # Save File
        f.save()

        # Update file thumbnail
        self.win.FileUpdated.emit(file_id)

    def __init__(self, model, *args):
        # Invoke parent init
        super().__init__(*args)

        # Get a reference to the window object
        self.win = get_app().window

        # Get Model data
        self.files_model = model
        self.setModel(self.files_model.proxy_model)

        # Remove the default selection model and wire up to the shared one
        self.selectionModel().deleteLater()
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionModel(self.files_model.selection_model)
        self.setSortingEnabled(True)
        # Keep "sortable" behavior available from the header, but do not apply
        # an initial forced sort so new imports keep insertion order by default.
        self.header().setSortIndicator(-1, Qt.AscendingOrder)
        self.files_model.proxy_model.sort(-1)
        self.setItemDelegate(FilesTreeProgressDelegate(self))

        self.setAcceptDrops(True)
        self.setDragEnabled(True)
        self.setDropIndicatorShown(True)

        # Setup header columns and layout
        self.setIconSize(info.TREE_ICON_SIZE)
        self.setIndentation(0)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.setWordWrap(False)
        self.setTextElideMode(Qt.ElideRight)

        self.files_model.ModelRefreshed.connect(self.refresh_view)

        # setup filter events
        self.files_model.model.itemChanged.connect(self.value_updated)



"""
 @file
 @brief Editable timecode field for the QWidget timeline ruler.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2025 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
"""

from qt_api import QColor, QLineEdit, Qt, pyqtSignal

from classes.time_parts import secondsToTime


def _color_css(color, fallback):
    col = QColor(color)
    if not col.isValid():
        col = QColor(fallback)
    return "rgba(%d, %d, %d, %.3f)" % (
        col.red(),
        col.green(),
        col.blue(),
        col.alpha() / 255.0,
    )


class TimecodeLineEdit(QLineEdit):
    """Small in-place timecode editor for the ruler playhead label."""

    frameCommitted = pyqtSignal(int, bool, bool)
    editCanceled = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self._fps_num = 30
        self._fps_den = 1
        self._current_frame = 1
        self._committing_focus_out = False
        self.setFrame(False)
        self.setObjectName("timelinePlayheadTimeEdit")
        self.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        self.setMaxLength(11)
        self.hide()

    def set_context(self, fps_num, fps_den, current_frame):
        self._fps_num = int(fps_num or 30)
        self._fps_den = int(fps_den or 1) or 1
        self._current_frame = max(1, int(current_frame or 1))

    def set_current_frame_text(self, frame):
        self._current_frame = max(1, int(frame or 1))
        self.setText(self.format_frame(self._current_frame))

    def apply_timeline_theme(self, theme, font, pad_left, pad_top):
        self.setFont(font)
        bg = theme.ruler_name_background
        if not bg.isValid():
            bg = theme.track.name_background
        fg = theme.ruler.font_color
        if not fg.isValid():
            fg = QColor("#ffffff")
        self.setTextMargins(max(0, int(pad_left)), 0, 2, 0)
        self.setStyleSheet(
            "QLineEdit#timelinePlayheadTimeEdit {"
            "background: %s;"
            "color: %s;"
            "border: 0px;"
            "border-radius: 0px;"
            "padding: 0px;"
            "}"
            % (
                _color_css(bg, "#000000"),
                _color_css(fg, "#ffffff"),
            )
        )

    def fps_float(self):
        if not self._fps_den:
            return 0.0
        return float(self._fps_num) / float(self._fps_den)

    def fps_frames(self):
        return max(1, int(self.fps_float() or 1.0))

    def format_frame(self, frame):
        fps = self.fps_float()
        seconds = 0.0
        if fps:
            seconds = max(0.0, (max(1, int(frame)) - 1) / fps)
        tt = secondsToTime(seconds, self._fps_num, self._fps_den)
        return "%s:%s:%s,%s" % (tt["hour"], tt["min"], tt["sec"], tt["frame"])

    def parse_text(self, text=None):
        raw = (self.text() if text is None else text).strip()
        if raw in ("", "0"):
            return 1
        raw = raw.replace(",", ":")
        parts = raw.split(":")
        if len(parts) != 4 or not all(part.strip().isdigit() for part in parts):
            return None
        hours, minutes, seconds, frames = [int(part.strip()) for part in parts]
        if minutes > 59 or seconds > 59 or frames >= self.fps_frames():
            return None
        fps = self.fps_float()
        if fps <= 0.0:
            return 1
        total_seconds = (hours * 3600) + (minutes * 60) + seconds
        return max(1, int(round(total_seconds * fps)) + frames + 1)

    def commit(self, start_preroll=True, force=True):
        frame = self.parse_text()
        if frame is None:
            self.editCanceled.emit()
            return False
        self.set_current_frame_text(frame)
        self.frameCommitted.emit(int(frame), bool(start_preroll), bool(force))
        return True

    def _segment_for_pos(self, pos):
        if pos <= 2:
            return "hour"
        if pos <= 5:
            return "minute"
        if pos <= 8:
            return "second"
        return "frame"

    def _segment_span(self, segment):
        return {
            "hour": (0, 2),
            "minute": (3, 2),
            "second": (6, 2),
            "frame": (9, 2),
        }.get(segment, (9, 2))

    def _step_segment(self, direction, modifiers):
        frame = self.parse_text()
        if frame is None:
            frame = self._current_frame

        segment = self._segment_for_pos(self.cursorPosition())
        fps = self.fps_float()
        fps_frames = self.fps_frames()
        if segment == "hour":
            delta = int(round(fps * 3600.0))
        elif segment == "minute":
            delta = int(round(fps * 60.0))
        elif segment == "second":
            delta = int(round(fps))
        else:
            delta = 1

        if modifiers & Qt.ShiftModifier:
            delta *= 10
        elif modifiers & Qt.ControlModifier and segment == "frame":
            delta = fps_frames

        frame = max(1, frame + (int(direction) * max(1, delta)))
        self.set_current_frame_text(frame)
        start, length = self._segment_span(segment)
        self.setSelection(start, length)
        self.frameCommitted.emit(int(frame), False, False)

    def keyPressEvent(self, event):
        key = event.key()
        if key in (Qt.Key_Return, Qt.Key_Enter):
            if self.commit(start_preroll=True, force=True):
                self._committing_focus_out = True
                try:
                    self.clearFocus()
                    self.hide()
                finally:
                    self._committing_focus_out = False
            event.accept()
            return
        if key == Qt.Key_Escape:
            self.editCanceled.emit()
            event.accept()
            return
        if key in (Qt.Key_Up, Qt.Key_Down):
            self._step_segment(1 if key == Qt.Key_Up else -1, event.modifiers())
            event.accept()
            return
        super().keyPressEvent(event)

    def focusOutEvent(self, event):
        if self.isVisible() and not self._committing_focus_out:
            self._committing_focus_out = True
            try:
                self.commit(start_preroll=True, force=True)
            finally:
                self._committing_focus_out = False
        super().focusOutEvent(event)



"""
 @file
 @brief This file contains the blender model, used by the 3d animated titles screen
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2018 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os

# Try to get the security-patched XML functions from defusedxml
try:
    from defusedxml import minidom as xml
except ImportError:
    from xml.dom import minidom as xml

from qt_api import Qt, QSortFilterProxyModel, QItemSelectionModel
from qt_api import QIcon, QStandardItem, QStandardItemModel

import openshot

from classes import info
from classes.logger import log
from classes.app import get_app


class BlenderFilterProxyModel(QSortFilterProxyModel):
    """Proxy class used for sorting and filtering model data"""

    def lessThan(self, left, right):
        """Sort blender model by a column at runtime"""
        leftData = left.data(self.sortRole())
        rightData = right.data(self.sortRole())

        return leftData < rightData


class BlenderModel():
    def update_model(self, clear=True):
        log.info("updating effects model.")

        _ = self.app._tr

        # Clear all items
        if clear:
            self.model_paths = {}
            self.model.clear()

        # Add Headers
        self.model.setHorizontalHeaderLabels([_("Thumb"), _("Name")])

        # get a list of files in the application blender directory
        blender_dir = os.path.join(info.PATH, "blender")
        icons_dir = os.path.join(blender_dir, "icons")

        for file in sorted(os.listdir(blender_dir)):
            path = os.path.join(blender_dir, file)
            if path in self.model_paths:
                continue
            if os.path.isfile(path) and ".xml" in file:
                # load xml effect file
                xmldoc = xml.parse(path)

                # Get column data for model
                title = xmldoc.getElementsByTagName("title")[0].childNodes[0].data
                icon_name = xmldoc.getElementsByTagName("icon")[0].childNodes[0].data
                icon_path = os.path.join(icons_dir, icon_name)
                service = xmldoc.getElementsByTagName("service")[0].childNodes[0].data
                xmldoc.unlink()

                # Check for thumbnail path (in build-in cache)
                thumb_path = os.path.join(info.IMAGES_PATH, "cache",  "blender_{}".format(icon_name))

                # Check built-in cache (if not found)
                if not os.path.exists(thumb_path):
                    # Check user folder cache
                    thumb_path = os.path.join(info.CACHE_PATH, "blender_{}".format(icon_name))

                # Check if thumb exists
                if not os.path.exists(thumb_path):

                    try:
                        # Reload this reader
                        clip = openshot.Clip(icon_path)
                        reader = clip.Reader()
                        reader.Open()

                        # Save thumbnail
                        reader.GetFrame(0).Thumbnail(
                            thumb_path, 98, 64, "", "",
                            "#000", False, "png", 85, 0.0)
                        reader.Close()
                    except Exception:
                        log.info('Invalid blender image file: %s', icon_path)
                        continue

                # Load icon (using display DPI)
                icon = QIcon()
                icon.addFile(thumb_path)

                row = []
                flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsUserCheckable
                # Append thumbnail
                col = QStandardItem(self.app._tr(title))
                col.setIcon(icon)
                col.setToolTip(self.app._tr(title))
                col.setFlags(flags)
                row.append(col)

                # Append Name
                col = QStandardItem(self.app._tr(title))
                col.setData(self.app._tr(title), Qt.DisplayRole)
                col.setFlags(flags)
                row.append(col)

                # Append Path
                col = QStandardItem(path)
                col.setData(path, Qt.DisplayRole)
                col.setFlags(flags)
                row.append(col)

                # Append Service
                col = QStandardItem(service)
                col.setData(service, Qt.DisplayRole)
                col.setFlags(flags)
                row.append(col)

                self.model.appendRow(row)
                self.model_paths[path] = path

                # Process events in QT (to keep the interface responsive)
                self.app.processEvents()

    def __init__(self, *args):

        # Create standard model
        self.app = get_app()
        self.model = QStandardItemModel()
        self.model.setColumnCount(3)
        self.model_paths = {}

        # Create proxy model (for sorting and filtering)
        self.proxy_model = BlenderFilterProxyModel()
        self.proxy_model.setDynamicSortFilter(True)
        self.proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.proxy_model.setFilterKeyColumn(1)
        self.proxy_model.setSortCaseSensitivity(Qt.CaseSensitive)
        self.proxy_model.setSourceModel(self.model)
        self.proxy_model.setSortLocaleAware(True)

        # Create selection model to share between views
        self.selection_model = QItemSelectionModel(self.proxy_model)
